from collections.abc import Sequence
from typing import assert_never

import click
from click_option_group import optgroup
from loguru import logger
from pydantic import Field

from imbue.imbue_common.frozen_model import FrozenModel
from imbue.mngr.api.data_types import GcResourceTypes
from imbue.mngr.api.gc import gc as api_gc
from imbue.mngr.api.list import load_all_agents_grouped_by_host
from imbue.mngr.api.providers import get_all_provider_instances
from imbue.mngr.api.providers import get_provider_instance
from imbue.mngr.cli.common_opts import CommonCliOptions
from imbue.mngr.cli.common_opts import add_common_options
from imbue.mngr.cli.common_opts import setup_command_context
from imbue.mngr.cli.completion import complete_agent_name
from imbue.mngr.cli.help_formatter import CommandHelpMetadata
from imbue.mngr.cli.help_formatter import add_pager_help_option
from imbue.mngr.cli.help_formatter import register_help_metadata
from imbue.mngr.cli.output_helpers import emit_event
from imbue.mngr.cli.output_helpers import emit_final_json
from imbue.mngr.cli.output_helpers import emit_format_template_lines
from imbue.mngr.cli.output_helpers import write_human_line
from imbue.mngr.config.data_types import MngrContext
from imbue.mngr.config.data_types import OutputOptions
from imbue.mngr.errors import AgentNotFoundError
from imbue.mngr.errors import HostOfflineError
from imbue.mngr.errors import MngrError
from imbue.mngr.errors import UserInputError
from imbue.mngr.interfaces.agent import AgentInterface
from imbue.mngr.interfaces.host import HostInterface
from imbue.mngr.interfaces.host import OnlineHostInterface
from imbue.mngr.interfaces.provider_instance import ProviderInstanceInterface
from imbue.mngr.primitives import AgentName
from imbue.mngr.primitives import ErrorBehavior
from imbue.mngr.primitives import OutputFormat


class _OfflineHostToDestroy(FrozenModel):
    """An offline host where all agents are targeted for destruction."""

    model_config = {**FrozenModel.model_config, "arbitrary_types_allowed": True}

    host: HostInterface = Field(description="The offline host to destroy")
    provider: ProviderInstanceInterface = Field(description="The provider instance for this host")
    agent_names: list[AgentName] = Field(description="Names of agents on this host targeted for destruction")


class _DestroyTargets(FrozenModel):
    """Result of finding agents/hosts to destroy."""

    model_config = {**FrozenModel.model_config, "arbitrary_types_allowed": True}

    online_agents: list[tuple[AgentInterface, OnlineHostInterface]] = Field(
        description="Agents on online hosts to destroy, paired with their host"
    )
    offline_hosts: list[_OfflineHostToDestroy] = Field(
        description="Offline hosts where all agents are targeted for destruction"
    )


def get_agent_name_from_session(session_name: str, prefix: str) -> str | None:
    """Extract the agent name from a tmux session name.

    The session name is expected to be in the format "{prefix}{agent_name}".
    Returns the agent name if the session matches the prefix, or None if the
    session name doesn't match the expected prefix format.
    """
    if not session_name:
        logger.debug("Failed to extract agent name: empty session name provided")
        return None

    # Check if the session name starts with our prefix
    if not session_name.startswith(prefix):
        logger.debug(
            "Failed to extract agent name: session name '{}' doesn't start with mngr prefix '{}'",
            session_name,
            prefix,
        )
        return None

    # Extract the agent name by removing the prefix
    agent_name = session_name[len(prefix) :]
    if not agent_name:
        logger.debug(
            "Failed to extract agent name: session name '{}' has empty agent name after stripping prefix", session_name
        )
        return None

    logger.debug("Extracted agent name '{}' from session '{}'", agent_name, session_name)
    return agent_name


class DestroyCliOptions(CommonCliOptions):
    """Options passed from the CLI to the destroy command.

    This captures all the click parameters so we can pass them as a single object
    to helper functions instead of passing dozens of individual parameters.

    Inherits common options (output_format, quiet, verbose, etc.) from CommonCliOptions.

    Note that this class VERY INTENTIONALLY DOES NOT use Field() decorators with descriptions, defaults, etc.
    For that information, see the click.option() and click.argument() decorators on the destroy() function itself.
    """

    agents: tuple[str, ...]
    agent_list: tuple[str, ...]
    force: bool
    destroy_all: bool
    dry_run: bool
    gc: bool
    sessions: tuple[str, ...]
    # Planned features (not yet implemented)
    include: tuple[str, ...]
    exclude: tuple[str, ...]
    stdin: bool


@click.command(name="destroy")
@click.argument("agents", nargs=-1, required=False, shell_complete=complete_agent_name)
@optgroup.group("Target Selection")
@optgroup.option(
    "--agent",
    "agent_list",
    multiple=True,
    help="Agent name or ID to destroy (can be specified multiple times)",
)
@optgroup.option(
    "-a",
    "--all",
    "--all-agents",
    "destroy_all",
    is_flag=True,
    help="Destroy all agents",
)
@optgroup.option(
    "--session",
    "sessions",
    multiple=True,
    help="Tmux session name to destroy (can be specified multiple times). The agent name is extracted by "
    "stripping the configured prefix from the session name.",
)
@optgroup.option(
    "--include",
    multiple=True,
    help="Filter agents to destroy by CEL expression (repeatable). [future]",
)
@optgroup.option(
    "--exclude",
    multiple=True,
    help="Exclude agents matching CEL expression from destruction (repeatable). [future]",
)
@optgroup.option(
    "--stdin",
    is_flag=True,
    help="Read agent names/IDs from stdin, one per line. [future]",
)
@optgroup.group("Behavior")
@optgroup.option(
    "-f",
    "--force",
    is_flag=True,
    help="Skip confirmation prompts and force destroy running agents",
)
@optgroup.option(
    "--dry-run",
    is_flag=True,
    help="Show what would be destroyed without actually destroying",
)
@optgroup.option(
    "--gc/--no-gc",
    default=True,
    help="Run garbage collection after destroying agents to clean up orphaned resources (default: enabled)",
)
@add_common_options
@click.pass_context
def destroy(ctx: click.Context, **kwargs) -> None:
    """Destroy agent(s) and clean up resources.

    When the last agent on a host is destroyed, the host itself is also destroyed.

    Use with caution! This operation is irreversible.

    Supports custom format templates via --format. Available fields: name.

    Examples:

      mngr destroy my-agent

      mngr destroy agent1 agent2 agent3

      mngr destroy --agent my-agent --agent another-agent

      mngr destroy --session mngr-my-agent

      mngr destroy --all --force

      mngr destroy --all --force --format '{name}'
    """
    # Setup command context (config, logging, output options)
    # This loads the config, applies defaults, and creates the final options
    mngr_ctx, output_opts, opts = setup_command_context(
        ctx=ctx,
        command_name="destroy",
        command_class=DestroyCliOptions,
        is_format_template_supported=True,
    )

    # Filter agents to destroy using CEL expressions like:
    # --include 'name.startsWith("test-")' or --include 'host.provider == "docker"'
    # See mngr list --include for the pattern to follow
    if opts.include:
        raise NotImplementedError(
            "The --include option is not yet implemented. "
            "See https://github.com/imbue-ai/mngr/issues/XXX for progress."
        )
    # Exclude agents matching CEL expressions from destruction:
    # --exclude 'state == "RUNNING"' to skip running agents
    # See mngr list --exclude for the pattern to follow
    if opts.exclude:
        raise NotImplementedError(
            "The --exclude option is not yet implemented. "
            "See https://github.com/imbue-ai/mngr/issues/XXX for progress."
        )
    # Read agent names/IDs from stdin to allow piping agent lists:
    # mngr list --format jsonl | jq -r .name | mngr destroy --stdin
    if opts.stdin:
        raise NotImplementedError(
            "The --stdin option is not yet implemented. See https://github.com/imbue-ai/mngr/issues/XXX for progress."
        )

    # Validate input
    agent_identifiers = list(opts.agents) + list(opts.agent_list)

    # Handle --session option by extracting agent names from session names
    if opts.sessions:
        if agent_identifiers or opts.destroy_all:
            raise UserInputError("Cannot specify --session with agent names or --all")
        for session_name in opts.sessions:
            agent_name = get_agent_name_from_session(session_name, mngr_ctx.config.prefix)
            if agent_name is None:
                raise UserInputError(
                    f"Session '{session_name}' does not match the expected format. "
                    f"Session names should start with the configured prefix '{mngr_ctx.config.prefix}'."
                )
            agent_identifiers.append(agent_name)

    if not agent_identifiers and not opts.destroy_all:
        raise UserInputError("Must specify at least one agent or use --all")

    if agent_identifiers and opts.destroy_all:
        raise UserInputError("Cannot specify both agent names and --all")

    # Find agents to destroy
    try:
        targets = _find_agents_to_destroy(
            agent_identifiers=agent_identifiers,
            destroy_all=opts.destroy_all,
            mngr_ctx=mngr_ctx,
        )
    except AgentNotFoundError as e:
        if opts.force:
            targets = _DestroyTargets(online_agents=[], offline_hosts=[])
            _output(f"Error destroying agent(s): {e}", output_opts)
        else:
            raise

    if not targets.online_agents and not targets.offline_hosts:
        _output("No agents found to destroy", output_opts)
        return

    # Handle dry-run mode
    if opts.dry_run:
        _output_targets(targets, "Would destroy:", output_opts)
        return

    # Confirm destruction if not forced
    if not opts.force:
        _confirm_destruction(targets)

    # Destroy agents on online hosts
    destroyed_agents: list[AgentName] = []
    for agent, host in targets.online_agents:
        try:
            if agent.is_running() and not opts.force:
                _output(
                    f"Agent {agent.name} is running. Use --force to destroy running agents.",
                    output_opts,
                )
                continue

            mngr_ctx.pm.hook.on_before_agent_destroy(agent=agent, host=host)
            host.destroy_agent(agent)
            mngr_ctx.pm.hook.on_agent_destroyed(agent=agent, host=host)
            destroyed_agents.append(agent.name)
            _output(f"Destroyed agent: {agent.name}", output_opts)

        except MngrError as e:
            _output(f"Error destroying agent {agent.name}: {e}", output_opts)

    # Destroy offline hosts (which destroys all their agents)
    for offline in targets.offline_hosts:
        try:
            _output(f"Destroying offline host with {len(offline.agent_names)} agent(s)...", output_opts)
            mngr_ctx.pm.hook.on_before_host_destroy(host=offline.host)
            offline.provider.destroy_host(offline.host)
            mngr_ctx.pm.hook.on_host_destroyed(host=offline.host)
            destroyed_agents.extend(offline.agent_names)
            for name in offline.agent_names:
                _output(f"Destroyed agent: {name} (via host destruction)", output_opts)
        except MngrError as e:
            _output(f"Error destroying offline host: {e}", output_opts)

    # Run garbage collection if enabled
    if opts.gc and not opts.dry_run and destroyed_agents:
        _run_post_destroy_gc(mngr_ctx=mngr_ctx, output_opts=output_opts)

    # Output final result
    _output_result(destroyed_agents, output_opts)


def _find_agents_to_destroy(
    agent_identifiers: Sequence[str],
    destroy_all: bool,
    mngr_ctx: MngrContext,
) -> _DestroyTargets:
    """Find all agents to destroy.

    Returns _DestroyTargets containing online agents and offline hosts to destroy.
    Raises AgentNotFoundError if any specified identifier does not match an agent.
    """
    online_agents: list[tuple[AgentInterface, OnlineHostInterface]] = []
    offline_hosts: list[_OfflineHostToDestroy] = []
    matched_identifiers: set[str] = set()
    seen_offline_hosts: set[str] = set()

    agents_by_host, _ = load_all_agents_grouped_by_host(mngr_ctx, include_destroyed=False)

    for host_ref, agent_refs in agents_by_host.items():
        for agent_ref in agent_refs:
            should_include: bool
            if destroy_all:
                should_include = True
            elif agent_identifiers:
                agent_name_str = str(agent_ref.agent_name)
                agent_id_str = str(agent_ref.agent_id)

                should_include = False
                for identifier in agent_identifiers:
                    if identifier == agent_name_str or identifier == agent_id_str:
                        should_include = True
                        matched_identifiers.add(identifier)
            else:
                should_include = False

            if should_include:
                provider = get_provider_instance(host_ref.provider_name, mngr_ctx)
                host_interface = provider.get_host(host_ref.host_id)

                match host_interface:
                    case OnlineHostInterface() as online_host:
                        for agent in online_host.get_agents():
                            if agent.id == agent_ref.agent_id:
                                online_agents.append((agent, online_host))
                                break
                        else:
                            raise AgentNotFoundError(
                                f"Agent with ID {agent_ref.agent_id} not found on host {online_host.id}"
                            )
                    case HostInterface() as offline_host:
                        host_id_str = str(host_ref.host_id)
                        if host_id_str in seen_offline_hosts:
                            continue

                        # Offline host - check if ALL agents on this host are being destroyed
                        all_agent_refs_on_host = offline_host.get_agent_references()
                        all_targeted = destroy_all or all(
                            str(ref.agent_name) in agent_identifiers or str(ref.agent_id) in agent_identifiers
                            for ref in all_agent_refs_on_host
                        )
                        if all_targeted:
                            # Collect the host for destruction (don't destroy yet)
                            offline_hosts.append(
                                _OfflineHostToDestroy(
                                    host=offline_host,
                                    provider=provider,
                                    agent_names=[ref.agent_name for ref in all_agent_refs_on_host],
                                )
                            )
                            seen_offline_hosts.add(host_id_str)
                            for ref in all_agent_refs_on_host:
                                matched_identifiers.add(str(ref.agent_name))
                                matched_identifiers.add(str(ref.agent_id))
                        else:
                            raise HostOfflineError(
                                f"Host '{host_ref.host_id}' is offline. Cannot destroy individual agents on an "
                                f"offline host. Either start the host first, or destroy all "
                                f"{len(all_agent_refs_on_host)} agent(s) on this host."
                            )
                    case _ as unreachable:
                        assert_never(unreachable)

    # Verify all specified identifiers were found
    if agent_identifiers:
        unmatched_identifiers = set(agent_identifiers) - matched_identifiers
        if unmatched_identifiers:
            unmatched_list = ", ".join(sorted(unmatched_identifiers))
            raise AgentNotFoundError(f"No agent(s) found matching: {unmatched_list}")

    return _DestroyTargets(online_agents=online_agents, offline_hosts=offline_hosts)


def _confirm_destruction(targets: _DestroyTargets) -> None:
    """Prompt user to confirm destruction of agents."""
    write_human_line("\nThe following agents will be destroyed:")
    for agent, _ in targets.online_agents:
        write_human_line("  - {}", agent.name)
    for offline in targets.offline_hosts:
        for name in offline.agent_names:
            write_human_line("  - {} (on offline host)", name)

    write_human_line("\nThis action is irreversible!")

    if not click.confirm("Are you sure you want to continue?"):
        raise click.Abort()


def _output_targets(
    targets: _DestroyTargets,
    prefix: str,
    output_opts: OutputOptions,
) -> None:
    """Output a list of agents to destroy."""
    agent_data = [
        {"agent_id": str(agent.id), "agent_name": str(agent.name), "host_id": str(host.id)}
        for agent, host in targets.online_agents
    ]
    for offline in targets.offline_hosts:
        for name in offline.agent_names:
            agent_data.append({"agent_name": str(name), "host_id": str(offline.host.id), "host_offline": True})

    match output_opts.output_format:
        case OutputFormat.JSON:
            emit_final_json({"agents": agent_data})
        case OutputFormat.JSONL:
            emit_event("agents_list", {"agents": agent_data}, OutputFormat.JSONL)
        case OutputFormat.HUMAN:
            write_human_line("\n{}", prefix)
            for agent, host in targets.online_agents:
                write_human_line("  - {} (on host {})", agent.name, host.id)
            for offline in targets.offline_hosts:
                for name in offline.agent_names:
                    write_human_line("  - {} (on offline host {})", name, offline.host.id)
        case _ as unreachable:
            assert_never(unreachable)


def _output(message: str, output_opts: OutputOptions) -> None:
    """Output a message according to the format."""
    if output_opts.output_format == OutputFormat.HUMAN:
        write_human_line(message)


def _output_result(destroyed_agents: Sequence[AgentName], output_opts: OutputOptions) -> None:
    """Output the final result."""
    if output_opts.format_template is not None:
        items = [{"name": str(n)} for n in destroyed_agents]
        emit_format_template_lines(output_opts.format_template, items)
        return
    result_data = {"destroyed_agents": [str(n) for n in destroyed_agents], "count": len(destroyed_agents)}
    match output_opts.output_format:
        case OutputFormat.JSON:
            emit_final_json(result_data)
        case OutputFormat.JSONL:
            emit_event("destroy_result", result_data, OutputFormat.JSONL)
        case OutputFormat.HUMAN:
            if destroyed_agents:
                write_human_line("\nSuccessfully destroyed {} agent(s)", len(destroyed_agents))
        case _ as unreachable:
            assert_never(unreachable)


def _run_post_destroy_gc(mngr_ctx: MngrContext, output_opts: OutputOptions) -> None:
    """Run garbage collection after destroying agents.

    This cleans up orphaned host-level resources (machines, work dirs, snapshots, volumes).
    Errors are logged but don't prevent destroy from reporting success.
    """
    try:
        _output("Garbage collecting...", output_opts)

        providers = get_all_provider_instances(mngr_ctx)

        resource_types = GcResourceTypes(
            is_machines=True,
            is_work_dirs=True,
            is_snapshots=True,
            is_volumes=True,
            is_logs=False,
            is_build_cache=False,
        )

        result = api_gc(
            mngr_ctx=mngr_ctx,
            providers=providers,
            resource_types=resource_types,
            include_filters=(),
            exclude_filters=(),
            dry_run=False,
            error_behavior=ErrorBehavior.CONTINUE,
        )

        _output("Garbage collecting... done.", output_opts)

        if result.errors:
            logger.warning("Garbage collection completed with {} error(s)", len(result.errors))
            for error in result.errors:
                logger.warning("  - {}", error)

    except MngrError as e:
        logger.warning("Garbage collection failed: {}", e)
        logger.warning("This does not affect the destroy operation, which completed successfully")


# Register help metadata for git-style help formatting
_DESTROY_HELP_METADATA = CommandHelpMetadata(
    name="mngr-destroy",
    one_line_description="Destroy agent(s) and clean up resources",
    synopsis="mngr [destroy|rm] [AGENTS...] [--agent <AGENT>] [--all] [--session <SESSION>] [-f|--force] [--dry-run]",
    description="""Destroy one or more agents and clean up their resources.

When the last agent on a host is destroyed, the host itself is also destroyed
(including containers, volumes, snapshots, and any remote infrastructure).

Use with caution! This operation is irreversible.

By default, running agents cannot be destroyed. Use --force to stop and destroy
running agents. The command will prompt for confirmation before destroying
agents unless --force is specified.""",
    aliases=("rm",),
    examples=(
        ("Destroy an agent by name", "mngr destroy my-agent"),
        ("Destroy multiple agents", "mngr destroy agent1 agent2 agent3"),
        ("Destroy all agents", "mngr destroy --all --force"),
        ("Preview what would be destroyed", "mngr destroy my-agent --dry-run"),
    ),
    see_also=(
        ("create", "Create a new agent"),
        ("list", "List existing agents"),
        ("gc", "Garbage collect orphaned resources"),
    ),
    additional_sections=(
        (
            "Related Documentation",
            """- [Resource Cleanup Options](../generic/resource_cleanup.md) - Control which associated resources are destroyed
- [Multi-target Options](../generic/multi_target.md) - Behavior when targeting multiple agents""",
        ),
    ),
)

register_help_metadata("destroy", _DESTROY_HELP_METADATA)
# Also register under alias for consistent help output
for alias in _DESTROY_HELP_METADATA.aliases:
    register_help_metadata(alias, _DESTROY_HELP_METADATA)

# Add pager-enabled help option to the destroy command
add_pager_help_option(destroy)



from __future__ import annotations

import platform
from datetime import datetime
from datetime import timezone
from typing import Final

from loguru import logger

from imbue.imbue_common.errors import SwitchError
from imbue.imbue_common.pure import pure
from imbue.mngr.config.data_types import MngrConfig
from imbue.mngr.primitives import ActivitySource
from imbue.mngr.primitives import AgentLifecycleState
from imbue.mngr.primitives import AgentTypeName
from imbue.mngr.primitives import CommandString
from imbue.mngr.primitives import IdleMode

LOCAL_CONNECTOR_NAME: Final[str] = "LocalConnector"


@pure
def is_macos() -> bool:
    """Check if the current system is macOS (Darwin)."""
    return platform.system() == "Darwin"


# Activity sources that are host-level (vs agent-level)
HOST_LEVEL_ACTIVITY_SOURCES: Final[frozenset[ActivitySource]] = frozenset(
    {
        ActivitySource.BOOT,
        ActivitySource.USER,
        ActivitySource.SSH,
    }
)


def get_activity_sources_for_idle_mode(idle_mode: IdleMode) -> tuple[ActivitySource, ...]:
    """Get the activity sources that should be monitored for a given idle mode.

    This mapping is defined in docs/concepts/idle_detection.md.
    """
    if idle_mode == IdleMode.IO:
        return (
            ActivitySource.USER,
            ActivitySource.AGENT,
            ActivitySource.SSH,
            ActivitySource.CREATE,
            ActivitySource.START,
            ActivitySource.BOOT,
        )
    elif idle_mode == IdleMode.USER:
        return (
            ActivitySource.USER,
            ActivitySource.SSH,
            ActivitySource.CREATE,
            ActivitySource.START,
            ActivitySource.BOOT,
        )
    elif idle_mode == IdleMode.AGENT:
        return (
            ActivitySource.AGENT,
            ActivitySource.SSH,
            ActivitySource.CREATE,
            ActivitySource.START,
            ActivitySource.BOOT,
        )
    elif idle_mode == IdleMode.SSH:
        return (
            ActivitySource.SSH,
            ActivitySource.CREATE,
            ActivitySource.START,
            ActivitySource.BOOT,
        )
    elif idle_mode == IdleMode.CREATE:
        return (ActivitySource.CREATE,)
    elif idle_mode == IdleMode.BOOT:
        return (ActivitySource.BOOT,)
    elif idle_mode == IdleMode.START:
        return (ActivitySource.START, ActivitySource.BOOT)
    elif idle_mode == IdleMode.RUN:
        return (
            ActivitySource.CREATE,
            ActivitySource.START,
            ActivitySource.BOOT,
            ActivitySource.PROCESS,
        )
    elif idle_mode == IdleMode.DISABLED:
        return ()
    else:
        raise SwitchError(idle_mode)


# =========================================================================
# Shared Listing Helpers
# =========================================================================

# Agent types that use a fixed expected process name instead of computing
# from the stored command. This handles agents like ClaudeAgent where the
# assembled command is a complex shell wrapper but the actual running
# process has a known name.
_EXPECTED_PROCESS_NAME_BY_AGENT_TYPE: Final[dict[str, str]] = {
    "claude": "claude",
}

# Common shell names for lifecycle state detection
SHELL_COMMANDS: Final[frozenset[str]] = frozenset({"bash", "sh", "zsh", "fish", "dash", "ksh", "tcsh", "csh"})


@pure
def resolve_expected_process_name(
    agent_type: str,
    command: CommandString,
    config: MngrConfig,
) -> str:
    """Resolve the expected process name for lifecycle state detection.

    For agent types with complex wrapper commands (like claude), returns the
    known process name. For custom types with a parent_type, resolves through
    the parent. Otherwise extracts the basename from the command.
    """
    # Resolve parent type for custom agent types
    effective_type = agent_type
    type_config = config.agent_types.get(AgentTypeName(agent_type))
    if type_config is not None and type_config.parent_type is not None:
        effective_type = str(type_config.parent_type)

    if effective_type in _EXPECTED_PROCESS_NAME_BY_AGENT_TYPE:
        return _EXPECTED_PROCESS_NAME_BY_AGENT_TYPE[effective_type]

    return command.split()[0].split("/")[-1] if command else ""


def compute_idle_seconds(
    user_activity: datetime | None,
    agent_activity: datetime | None,
    ssh_activity: datetime | None,
) -> float | None:
    """Compute idle seconds from the most recent activity time."""
    latest_activity: datetime | None = None
    for activity_time in (user_activity, agent_activity, ssh_activity):
        if activity_time is not None:
            if latest_activity is None or activity_time > latest_activity:
                latest_activity = activity_time
    if latest_activity is None:
        return None
    return (datetime.now(timezone.utc) - latest_activity).total_seconds()


@pure
def timestamp_to_datetime(timestamp: int | None) -> datetime | None:
    """Convert a Unix timestamp to a UTC datetime, or None if the timestamp is None."""
    if timestamp is None:
        return None
    try:
        return datetime.fromtimestamp(timestamp, tz=timezone.utc)
    except (ValueError, OSError) as e:
        logger.trace("Failed to convert timestamp {} to datetime: {}", timestamp, e)
        return None


@pure
def get_descendant_process_names(root_pid: str, ps_output: str) -> list[str]:
    """Get names of all descendant processes from ps output."""
    children_by_ppid: dict[str, list[str]] = {}
    comm_by_pid: dict[str, str] = {}

    for line in ps_output.strip().split("\n"):
        line_parts = line.split()
        if len(line_parts) >= 3:
            pid, ppid, comm = line_parts[0], line_parts[1], line_parts[2]
            comm_by_pid[pid] = comm
            if ppid not in children_by_ppid:
                children_by_ppid[ppid] = []
            children_by_ppid[ppid].append(pid)

    descendant_names: list[str] = []
    queue = list(children_by_ppid.get(root_pid, []))
    while queue:
        pid = queue.pop(0)
        if pid in comm_by_pid:
            descendant_names.append(comm_by_pid[pid])
        queue.extend(children_by_ppid.get(pid, []))

    return descendant_names


@pure
def determine_lifecycle_state(
    tmux_info: str | None,
    is_active: bool,
    expected_process_name: str,
    ps_output: str,
) -> AgentLifecycleState:
    """Determine agent lifecycle state from tmux info and ps output.

    This is a pure function that replicates the logic from
    BaseAgent.get_lifecycle_state() using pre-collected data instead of
    making SSH calls.
    """
    if not tmux_info:
        return AgentLifecycleState.STOPPED

    parts = tmux_info.split("|")
    if len(parts) != 3:
        return AgentLifecycleState.STOPPED

    pane_dead, current_command, pane_pid = parts

    if pane_dead == "1":
        return AgentLifecycleState.DONE

    if current_command == expected_process_name:
        return AgentLifecycleState.RUNNING if is_active else AgentLifecycleState.WAITING

    # Check descendant processes
    descendant_names = get_descendant_process_names(pane_pid, ps_output)

    if expected_process_name in descendant_names:
        return AgentLifecycleState.RUNNING if is_active else AgentLifecycleState.WAITING

    # Check for non-shell descendant processes
    non_shell_processes = [p for p in descendant_names if p not in SHELL_COMMANDS]
    if non_shell_processes:
        return AgentLifecycleState.REPLACED

    # Current command is a shell -> agent probably finished
    if current_command in SHELL_COMMANDS:
        return AgentLifecycleState.DONE

    return AgentLifecycleState.REPLACED



"""Release tests for the claude agent provisioning on Modal.

These tests require Modal credentials, network access, and Claude credentials
to run. They are marked with @pytest.mark.release and only run when pushing
to main. To run them locally:

    PYTEST_MAX_DURATION=600 uv run pytest --no-cov --cov-fail-under=0 -n 0 -m release \\
        libs/mngr/imbue/mngr/agents/default_plugins/test_claude_agent.py::test_claude_agent_provisioning_on_modal
"""

import subprocess
from pathlib import Path
from uuid import uuid4

import pytest

from imbue.mngr.conftest import ModalSubprocessTestEnv


@pytest.fixture
def temp_source_dir(tmp_path: Path) -> Path:
    """Create a temporary source directory for tests."""
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    # Create a simple file so the directory isn't empty
    (source_dir / "test.txt").write_text("test content")
    return source_dir


@pytest.mark.release
@pytest.mark.timeout(600)
def test_claude_agent_provisioning_on_modal(
    temp_source_dir: Path,
    modal_subprocess_env: ModalSubprocessTestEnv,
) -> None:
    """Test creating a claude agent on Modal.

    This is an end-to-end release test that verifies:
    1. Claude agent can be provisioned on Modal
    2. Claude credentials are transferred correctly (if available locally)
    3. Claude is installed on the remote host
    4. The agent is created and started successfully

    The test uses --dangerously-skip-permissions -p "just say 'hello'" to run
    a quick, non-interactive claude session. The actual output goes to tmux,
    so we only verify that the agent was created successfully.
    """
    # Use a unique agent name with globally unique id to avoid collisions
    unique_id = uuid4().hex[:12]
    agent_name = f"test-claude-modal-{unique_id}"

    # make a .gitignore file to ignore the claude local settings
    claude_settings_dir = temp_source_dir / ".claude"
    claude_settings_dir.mkdir()
    (claude_settings_dir / "settings.local.json").write_text("{}")
    (temp_source_dir / ".gitignore").write_text(".claude/settings.local.json\n")

    # Run mngr create with claude agent on modal
    # Using --no-connect and --await-ready to run synchronously without attaching
    # Using --no-ensure-clean since temp dir won't be a git repo
    result = subprocess.run(
        [
            "uv",
            "run",
            "mngr",
            "create",
            agent_name,
            "claude",
            "--in",
            "modal",
            "--no-connect",
            "--await-ready",
            "--no-ensure-clean",
            "--source",
            str(temp_source_dir),
            "--",
            "--dangerously-skip-permissions",
            "-p",
            "just say 'hello'",
        ],
        capture_output=True,
        text=True,
        timeout=600,
        env=modal_subprocess_env.env,
    )

    # Check that the command succeeded
    assert result.returncode == 0, f"CLI failed with stderr: {result.stderr}\nstdout: {result.stdout}"
    assert "Done." in result.stdout, f"Expected 'Done.' in output: {result.stdout}"

    # Verify that Claude was installed (this message appears in the provisioning output)
    # This confirms that the claude plugin provisioning hook ran correctly
    combined_output = result.stdout + result.stderr
    assert "Claude installed successfully" in combined_output or "Claude is already installed" in combined_output, (
        f"Expected Claude installation message in output.\nstdout: {result.stdout}\nstderr: {result.stderr}"
    )



"""Tests for the LocalProviderBackend."""

import os
from pathlib import Path

import pluggy

from imbue.mngr.config.data_types import MngrConfig
from imbue.mngr.config.data_types import MngrContext
from imbue.mngr.primitives import HostName
from imbue.mngr.primitives import ProviderBackendName
from imbue.mngr.primitives import ProviderInstanceName
from imbue.mngr.providers.local.backend import LOCAL_BACKEND_NAME
from imbue.mngr.providers.local.backend import LocalProviderBackend
from imbue.mngr.providers.local.config import LocalProviderConfig
from imbue.mngr.providers.local.instance import LocalProviderInstance


def test_backend_name() -> None:
    assert LocalProviderBackend.get_name() == LOCAL_BACKEND_NAME
    assert LocalProviderBackend.get_name() == ProviderBackendName("local")


def test_backend_description() -> None:
    assert "local" in LocalProviderBackend.get_description().lower()


def test_backend_build_args_help() -> None:
    help_text = LocalProviderBackend.get_build_args_help()
    assert isinstance(help_text, str)
    assert len(help_text) > 0


def test_backend_start_args_help() -> None:
    help_text = LocalProviderBackend.get_start_args_help()
    assert isinstance(help_text, str)
    assert len(help_text) > 0


def test_backend_get_config_class() -> None:
    assert LocalProviderBackend.get_config_class() is LocalProviderConfig


def test_build_provider_instance_returns_local_provider_instance(temp_mngr_ctx: MngrContext) -> None:
    config = LocalProviderConfig()
    instance = LocalProviderBackend.build_provider_instance(
        name=ProviderInstanceName("test"),
        config=config,
        mngr_ctx=temp_mngr_ctx,
    )
    assert isinstance(instance, LocalProviderInstance)


def test_build_provider_instance_with_custom_host_dir(tmp_path: Path, temp_mngr_ctx: MngrContext) -> None:
    custom_dir = tmp_path / "custom_host_dir"
    custom_dir.mkdir()
    config = LocalProviderConfig(host_dir=custom_dir)
    instance = LocalProviderBackend.build_provider_instance(
        name=ProviderInstanceName("test"),
        config=config,
        mngr_ctx=temp_mngr_ctx,
    )
    assert isinstance(instance, LocalProviderInstance)
    # host_dir should be the custom_dir directly
    assert instance.host_dir == custom_dir


def test_build_provider_instance_uses_default_host_dir(temp_mngr_ctx: MngrContext) -> None:
    config = LocalProviderConfig()
    instance = LocalProviderBackend.build_provider_instance(
        name=ProviderInstanceName("test"),
        config=config,
        mngr_ctx=temp_mngr_ctx,
    )
    # host_dir should be the expanded default_host_dir directly
    expanded_default = Path(os.path.expanduser(str(temp_mngr_ctx.config.default_host_dir)))
    assert instance.host_dir == expanded_default


def test_build_provider_instance_uses_config_default_host_dir(temp_mngr_ctx: MngrContext) -> None:
    config = LocalProviderConfig()
    instance = LocalProviderBackend.build_provider_instance(
        name=ProviderInstanceName("test"),
        config=config,
        mngr_ctx=temp_mngr_ctx,
    )
    # host_dir should equal the expanded default_host_dir
    assert instance.host_dir == temp_mngr_ctx.config.default_host_dir.expanduser()


def test_build_provider_instance_uses_name(temp_mngr_ctx: MngrContext) -> None:
    config = LocalProviderConfig()
    instance = LocalProviderBackend.build_provider_instance(
        name=ProviderInstanceName("my-local"),
        config=config,
        mngr_ctx=temp_mngr_ctx,
    )
    assert instance.name == ProviderInstanceName("my-local")


def test_built_instance_can_create_host(tmp_path: Path, temp_mngr_ctx: MngrContext) -> None:
    custom_dir = tmp_path / "host_dir"
    custom_dir.mkdir()
    config = LocalProviderConfig(host_dir=custom_dir)
    instance = LocalProviderBackend.build_provider_instance(
        name=ProviderInstanceName("test"),
        config=config,
        mngr_ctx=temp_mngr_ctx,
    )

    host = instance.create_host(HostName("localhost"))
    assert host is not None
    assert host.id is not None


def test_multiple_instances_with_different_names(
    tmp_path: Path,
    temp_profile_dir: Path,
    mngr_test_prefix: str,
    plugin_manager: pluggy.PluginManager,
) -> None:
    # Create two separate host directories
    tmpdir1 = tmp_path / "host1"
    tmpdir2 = tmp_path / "host2"
    tmpdir1.mkdir()
    tmpdir2.mkdir()

    mngr_ctx1 = MngrContext(
        config=MngrConfig(default_host_dir=tmpdir1, prefix=mngr_test_prefix),
        pm=plugin_manager,
        profile_dir=temp_profile_dir,
    )
    mngr_ctx2 = MngrContext(
        config=MngrConfig(default_host_dir=tmpdir2, prefix=mngr_test_prefix),
        pm=plugin_manager,
        profile_dir=temp_profile_dir,
    )
    config1 = LocalProviderConfig(host_dir=tmpdir1)
    config2 = LocalProviderConfig(host_dir=tmpdir2)
    instance1 = LocalProviderBackend.build_provider_instance(
        name=ProviderInstanceName("local-1"),
        config=config1,
        mngr_ctx=mngr_ctx1,
    )
    instance2 = LocalProviderBackend.build_provider_instance(
        name=ProviderInstanceName("local-2"),
        config=config2,
        mngr_ctx=mngr_ctx2,
    )

    assert instance1.name == ProviderInstanceName("local-1")
    assert instance2.name == ProviderInstanceName("local-2")

    host1 = instance1.create_host(HostName("localhost"))
    host2 = instance2.create_host(HostName("localhost"))

    assert host1.id != host2.id



import sys
import time
from typing import Any

import click
from click_option_group import optgroup
from loguru import logger
from pydantic import ConfigDict
from urwid.display.raw import Screen
from urwid.event_loop.abstract_loop import ExitMainLoop
from urwid.event_loop.main_loop import MainLoop
from urwid.widget.attr_map import AttrMap
from urwid.widget.divider import Divider
from urwid.widget.frame import Frame
from urwid.widget.listbox import ListBox
from urwid.widget.listbox import SimpleFocusListWalker
from urwid.widget.pile import Pile
from urwid.widget.text import Text
from urwid.widget.wimp import SelectableIcon

from imbue.imbue_common.mutable_model import MutableModel
from imbue.imbue_common.pure import pure
from imbue.mngr.api.connect import connect_to_agent
from imbue.mngr.api.data_types import ConnectionOptions
from imbue.mngr.api.find import find_and_maybe_start_agent_by_name_or_id
from imbue.mngr.api.list import list_agents
from imbue.mngr.api.list import load_all_agents_grouped_by_host
from imbue.mngr.cli.common_opts import CommonCliOptions
from imbue.mngr.cli.common_opts import add_common_options
from imbue.mngr.cli.common_opts import setup_command_context
from imbue.mngr.cli.completion import complete_agent_name
from imbue.mngr.cli.help_formatter import CommandHelpMetadata
from imbue.mngr.cli.help_formatter import add_pager_help_option
from imbue.mngr.cli.help_formatter import register_help_metadata
from imbue.mngr.errors import UserInputError
from imbue.mngr.interfaces.agent import AgentInterface
from imbue.mngr.interfaces.data_types import AgentInfo
from imbue.mngr.interfaces.host import DEFAULT_AGENT_READY_TIMEOUT_SECONDS
from imbue.mngr.interfaces.host import OnlineHostInterface
from imbue.mngr.primitives import AgentLifecycleState


class ConnectCliOptions(CommonCliOptions):
    """Options passed from the CLI to the connect command.

    Inherits common options (output_format, quiet, verbose, etc.) from CommonCliOptions.
    """

    agent: str | None
    start: bool
    reconnect: bool
    message: str | None
    message_file: str | None
    ready_timeout: float
    retry: int
    retry_delay: str
    attach_command: str | None
    allow_unknown_host: bool


@pure
def filter_agents(
    agents: list[AgentInfo],
    hide_stopped: bool,
    search_query: str,
) -> list[AgentInfo]:
    """Filter agents by stopped state and search query."""
    result = agents

    if hide_stopped:
        result = [a for a in result if a.state != AgentLifecycleState.STOPPED]

    if search_query:
        query_lower = search_query.lower()
        result = [a for a in result if query_lower in str(a.name).lower()]

    return result


def build_status_text(
    search_query: str,
    hide_stopped: bool,
) -> str:
    """Build the status bar text for the agent selector."""
    parts = ["Status: Ready"]

    if search_query:
        parts.append(f"Search: {search_query}")
    else:
        parts.append("Type to search")

    if hide_stopped:
        parts.append("Filter: Hiding stopped")
    else:
        parts.append("Filter: All agents")

    return " | ".join(parts)


def handle_search_key(
    key: str,
    is_printable: bool,
    character: str | None,
    current_query: str,
) -> tuple[str, bool]:
    """Handle a key press for typeahead search. Returns (new_query, should_refresh)."""
    if key == "backspace":
        if current_query:
            return current_query[:-1], True
        else:
            return current_query, False
    elif is_printable and character:
        return current_query + character, True
    else:
        return current_query, False


def _create_selectable_agent_item(agent: AgentInfo, name_width: int, state_width: int) -> AttrMap:
    """Create a selectable list item representing an agent as a table row.

    Uses SelectableIcon instead of Text so that ListBox can navigate between items.
    urwid.Text is not selectable, which prevents ListBox arrow key navigation.
    """
    # Pad the name and state to their column widths for proper alignment
    name_padded = str(agent.name).ljust(name_width)
    state_padded = agent.state.value.ljust(state_width)
    host_str = str(agent.host.name)

    # Create a single SelectableIcon with the full formatted row
    # This ensures the entire row is selectable as one unit
    display_text = f"{name_padded}  {state_padded}  {host_str}"
    selectable_item = SelectableIcon(display_text, cursor_position=0)

    return AttrMap(selectable_item, None, focus_map="reversed")


class AgentSelectorState(MutableModel):
    """Mutable state for the agent selector UI."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    agents: list[AgentInfo]
    filtered_agents: list[AgentInfo] = []
    list_walker: Any
    status_text: Any
    result: AgentInfo | None = None
    hide_stopped: bool = False
    search_query: str = ""
    last_ctrl_c_time: float = 0.0
    name_width: int = 0
    state_width: int = 0


def _refresh_agent_list(state: AgentSelectorState) -> None:
    """Refresh the agent list view with current filter settings."""
    state.filtered_agents = filter_agents(state.agents, state.hide_stopped, state.search_query)

    state.list_walker.clear()
    for agent in state.filtered_agents:
        state.list_walker.append(_create_selectable_agent_item(agent, state.name_width, state.state_width))

    if state.list_walker:
        state.list_walker.set_focus(0)

    state.status_text.set_text(build_status_text(state.search_query, state.hide_stopped))


def _handle_selector_input(state: AgentSelectorState, key: str) -> bool:
    """Handle keyboard input for the agent selector. Returns True if handled, False to pass through."""
    if key == "ctrl r":
        state.hide_stopped = not state.hide_stopped
        _refresh_agent_list(state)
        return True

    if key == "ctrl c":
        current_time = time.time()
        if state.search_query:
            # First Ctrl-c clears the search query
            state.search_query = ""
            state.last_ctrl_c_time = current_time
            _refresh_agent_list(state)
            return True
        elif current_time - state.last_ctrl_c_time < 0.5:
            # Second Ctrl-c within 500ms exits
            raise ExitMainLoop()
        else:
            # Single Ctrl-c with no query - record time and wait for potential second
            state.last_ctrl_c_time = current_time
            return True

    if key == "enter":
        if state.list_walker and state.filtered_agents:
            _, focus_index = state.list_walker.get_focus()
            if focus_index is not None and 0 <= focus_index < len(state.filtered_agents):
                state.result = state.filtered_agents[focus_index]
        raise ExitMainLoop()

    # Let arrow keys pass through to the ListBox for navigation
    if key in ("up", "down", "page up", "page down", "home", "end"):
        return False

    is_printable = len(key) == 1 and key.isprintable()
    character = key if is_printable else None

    new_query, should_refresh = handle_search_key(
        key=key,
        is_printable=is_printable,
        character=character,
        current_query=state.search_query,
    )

    if should_refresh:
        state.search_query = new_query
        _refresh_agent_list(state)
        return True

    return False


class SelectorInputHandler(MutableModel):
    """Callable input handler for urwid MainLoop."""

    state: AgentSelectorState

    def __call__(self, key: str | tuple[str, int, int, int]) -> bool | None:
        """Handle keyboard input. Returns True if handled, None to pass through."""
        if isinstance(key, tuple):
            return None
        handled = _handle_selector_input(self.state, key)
        return True if handled else None


def _run_agent_selector(agents: list[AgentInfo]) -> AgentInfo | None:
    """Run the agent selector UI and return the selected agent, or None if cancelled."""
    # Calculate column widths based on content
    name_width = max((len(str(a.name)) for a in agents), default=10)
    state_width = max((len(a.state.value) for a in agents), default=7)

    # Cap widths at reasonable maximums
    name_width = min(name_width, 40)
    state_width = min(state_width, 15)

    list_walker: SimpleFocusListWalker[AttrMap] = SimpleFocusListWalker([])
    listbox = ListBox(list_walker)

    status_text = Text(build_status_text("", False))
    status_bar = AttrMap(status_text, "status")

    state = AgentSelectorState(
        agents=agents,
        list_walker=list_walker,
        status_text=status_text,
        name_width=name_width,
        state_width=state_width,
    )

    instructions_text = (
        "Instructions:\n"
        "  Type - Search agents by name\n"
        "  Up/Down - Navigate the list\n"
        "  Enter - Select an agent\n"
        "  Backspace - Clear search character\n"
        "  Ctrl+C - Clear search (twice to quit)\n"
        "  Ctrl+R - Toggle hiding stopped agents"
    )
    instructions = Text(instructions_text)

    # Create table header matching the SelectableIcon format in list items
    header_text = f"{'NAME'.ljust(name_width)}  {'STATE'.ljust(state_width)}  HOST"
    header_row = AttrMap(Text(("table_header", header_text)), "table_header")

    _refresh_agent_list(state)

    header = Pile(
        [
            AttrMap(Text("Agent Selector", align="center"), "header"),
            Divider(),
            instructions,
            Divider(),
            header_row,
            Divider("-"),
        ]
    )

    footer = Pile(
        [
            Divider(),
            status_bar,
        ]
    )

    frame = Frame(
        body=listbox,
        header=header,
        footer=footer,
    )

    palette = [
        ("header", "white", "dark blue"),
        ("status", "white", "dark blue"),
        ("reversed", "standout", ""),
        ("table_header", "bold", ""),
    ]

    input_handler = SelectorInputHandler(state=state)

    # Create screen and disable Ctrl-c SIGINT mapping so we can handle it as a key
    screen = Screen()
    screen.tty_signal_keys(intr="undefined")

    loop = MainLoop(
        frame,
        palette=palette,
        unhandled_input=input_handler,
        screen=screen,
    )
    loop.run()

    return state.result


def select_agent_interactively(agents: list[AgentInfo]) -> AgentInfo | None:
    """Show an interactive UI to select an agent. Returns None if cancelled."""
    if not agents:
        return None

    return _run_agent_selector(agents)


@pure
def _build_connection_options(opts: ConnectCliOptions) -> ConnectionOptions:
    """Build ConnectionOptions from CLI options."""
    return ConnectionOptions(
        is_reconnect=opts.reconnect,
        retry_count=opts.retry,
        retry_delay=opts.retry_delay,
        attach_command=opts.attach_command,
        is_unknown_host_allowed=opts.allow_unknown_host,
    )


@click.command()
@click.argument("agent", default=None, required=False, shell_complete=complete_agent_name)
@optgroup.group("General")
@optgroup.option("--agent", "agent", help="The agent to connect to (by name or ID)")
@optgroup.option(
    "--start/--no-start",
    default=True,
    show_default=True,
    help="Automatically start the agent if stopped",
)
@optgroup.group("Options")
@optgroup.option(
    "--reconnect/--no-reconnect",
    default=True,
    show_default=True,
    help="Automatically reconnect if dropped [future]",
)
@optgroup.option("--message", help="Initial message to send after connecting [future]")
@optgroup.option(
    "--message-file", type=click.Path(exists=True), help="File containing initial message to send [future]"
)
@optgroup.option(
    "--ready-timeout",
    type=float,
    default=DEFAULT_AGENT_READY_TIMEOUT_SECONDS,
    show_default=True,
    help="Timeout in seconds to wait for agent readiness [future]",
)
@optgroup.option("--retry", type=int, default=3, show_default=True, help="Number of connection retries [future]")
@optgroup.option("--retry-delay", default="5s", show_default=True, help="Delay between retries [future]")
@optgroup.option("--attach-command", help="Command to run instead of attaching to main session [future]")
@optgroup.option(
    "--allow-unknown-host/--no-allow-unknown-host",
    "allow_unknown_host",
    default=False,
    show_default=True,
    help="Allow connecting to hosts without a known_hosts file (disables SSH host key verification)",
)
@add_common_options
@click.pass_context
def connect(ctx: click.Context, **kwargs: Any) -> None:
    """Connect to an existing agent via the terminal.

    Attaches to the agent's tmux session, roughly equivalent to SSH'ing into
    the agent's machine and attaching to the tmux session. Use `mngr open` to
    open an agent's URLs in a web browser instead.

    If no agent is specified, shows an interactive selector to choose from
    available agents.

    \b
    Alias: conn
    """
    mngr_ctx, output_opts, opts = setup_command_context(
        ctx=ctx,
        command_name="connect",
        command_class=ConnectCliOptions,
    )

    # Send the specified text as an initial message after the agent starts
    # Should wait for ready_timeout seconds for agent readiness before sending
    if opts.message is not None:
        raise NotImplementedError("--message is not implemented yet")

    # Read initial message content from the specified file and send after agent starts
    if opts.message_file is not None:
        raise NotImplementedError("--message-file is not implemented yet")

    # Timeout for waiting for agent readiness
    if opts.ready_timeout != DEFAULT_AGENT_READY_TIMEOUT_SECONDS:
        raise NotImplementedError("--ready-timeout with non-default value is not implemented yet")

    # Number of times to retry connection on failure before giving up
    if opts.retry != 3:
        raise NotImplementedError("--retry with non-default value is not implemented yet")

    # Delay between connection retries (supports durations like "5s", "1m")
    if opts.retry_delay != "5s":
        raise NotImplementedError("--retry-delay with non-default value is not implemented yet")

    # Run this command instead of the default tmux attach
    # Useful for running a different shell or command in the agent's environment
    if opts.attach_command is not None:
        raise NotImplementedError("--attach-command is not implemented yet")

    # Disable automatic reconnection if the connection is dropped
    # Default behavior (--reconnect) should automatically reconnect
    if not opts.reconnect:
        raise NotImplementedError("--no-reconnect is not implemented yet")

    logger.info("Finding agent...")
    agents_by_host, providers = load_all_agents_grouped_by_host(mngr_ctx)

    agent: AgentInterface
    host: OnlineHostInterface

    if opts.agent is not None:
        agent, host = find_and_maybe_start_agent_by_name_or_id(
            opts.agent,
            agents_by_host,
            mngr_ctx,
            "connect",
            is_start_desired=opts.start,
        )
    elif not sys.stdin.isatty():
        # Default to most recently created agent when running non-interactively
        list_result = list_agents(mngr_ctx, is_streaming=False)
        if not list_result.agents:
            raise UserInputError("No agents found")

        # Sort by create_time descending to get most recent first
        sorted_agents = sorted(list_result.agents, key=lambda a: a.create_time, reverse=True)
        most_recent = sorted_agents[0]
        logger.info("No agent specified, connecting to most recently created: {}", most_recent.name)
        agent, host = find_and_maybe_start_agent_by_name_or_id(
            str(most_recent.id),
            agents_by_host,
            mngr_ctx,
            "connect",
            is_start_desired=opts.start,
        )
    else:
        list_result = list_agents(mngr_ctx, is_streaming=False)
        if not list_result.agents:
            raise UserInputError("No agents found")

        selected = select_agent_interactively(list_result.agents)
        if selected is None:
            logger.info("No agent selected")
            return

        agent, host = find_and_maybe_start_agent_by_name_or_id(
            str(selected.id),
            agents_by_host,
            mngr_ctx,
            "connect",
            is_start_desired=opts.start,
        )

    # Build connection options
    connection_opts = _build_connection_options(opts)

    logger.info("Connecting to agent: {}", agent.name)
    connect_to_agent(agent, host, mngr_ctx, connection_opts)


# Register help metadata for git-style help formatting
_CONNECT_HELP_METADATA = CommandHelpMetadata(
    name="mngr-connect",
    one_line_description="Connect to an existing agent via the terminal",
    synopsis="mngr [connect|conn] [OPTIONS] [AGENT]",
    description="""Connect to an existing agent via the terminal.

Attaches to the agent's tmux session, roughly equivalent to SSH'ing into
the agent's machine and attaching to the tmux session. Use `mngr open` to
open an agent's URLs in a web browser instead.

If no agent is specified, shows an interactive selector to choose from
available agents. The selector allows typeahead search to filter agents
by name.

The agent can be specified as a positional argument or via --agent:
  mngr connect my-agent
  mngr connect --agent my-agent""",
    aliases=("conn",),
    examples=(
        ("Connect to an agent by name", "mngr connect my-agent"),
        ("Connect without auto-starting if stopped", "mngr connect my-agent --no-start"),
        ("Show interactive agent selector", "mngr connect"),
    ),
    see_also=(
        ("create", "Create and connect to a new agent"),
        ("list", "List available agents"),
    ),
)

register_help_metadata("connect", _CONNECT_HELP_METADATA)
# Also register under alias for consistent help output
for alias in _CONNECT_HELP_METADATA.aliases:
    register_help_metadata(alias, _CONNECT_HELP_METADATA)

# Add pager-enabled help option to the connect command
add_pager_help_option(connect)



"""Tests for the connect CLI command."""

import time
from pathlib import Path

import pluggy
import pytest
from click.testing import CliRunner
from urwid.event_loop.abstract_loop import ExitMainLoop
from urwid.widget.attr_map import AttrMap
from urwid.widget.listbox import SimpleFocusListWalker
from urwid.widget.text import Text
from urwid.widget.wimp import SelectableIcon

from imbue.imbue_common.model_update import to_update
from imbue.mngr.cli.conftest import make_test_agent_info
from imbue.mngr.cli.connect import AgentSelectorState
from imbue.mngr.cli.connect import ConnectCliOptions
from imbue.mngr.cli.connect import SelectorInputHandler
from imbue.mngr.cli.connect import _build_connection_options
from imbue.mngr.cli.connect import _create_selectable_agent_item
from imbue.mngr.cli.connect import _handle_selector_input
from imbue.mngr.cli.connect import _refresh_agent_list
from imbue.mngr.cli.connect import build_status_text
from imbue.mngr.cli.connect import connect
from imbue.mngr.cli.connect import filter_agents
from imbue.mngr.cli.connect import handle_search_key
from imbue.mngr.cli.connect import select_agent_interactively
from imbue.mngr.cli.create import create
from imbue.mngr.interfaces.data_types import AgentInfo
from imbue.mngr.main import cli
from imbue.mngr.primitives import AgentLifecycleState
from imbue.mngr.primitives import AgentName
from imbue.mngr.utils.testing import cleanup_tmux_session
from imbue.mngr.utils.testing import create_test_agent_via_cli
from imbue.mngr.utils.testing import tmux_session_cleanup
from imbue.mngr.utils.testing import tmux_session_exists

# =============================================================================
# CLI-level integration tests for connect command
#
# These tests invoke the connect CLI command end-to-end. Because os.execvp
# replaces the test process, we intercept it via the intercepted_execvp_calls
# fixture (defined in conftest.py) to capture the args and verify the full
# CLI pipeline works.
# =============================================================================


def test_connect_no_agent_found(
    cli_runner: CliRunner,
    temp_work_dir: Path,
    plugin_manager: pluggy.PluginManager,
) -> None:
    """Test that connecting to a non-existent agent raises an error."""
    result = cli_runner.invoke(
        connect,
        ["nonexistent-agent-xyz123"],
        obj=plugin_manager,
    )

    assert result.exit_code != 0
    assert result.exception is not None


def test_connect_cli_invokes_tmux_attach_for_named_agent(
    cli_runner: CliRunner,
    temp_work_dir: Path,
    mngr_test_prefix: str,
    plugin_manager: pluggy.PluginManager,
    intercepted_execvp_calls: list[tuple[str, list[str]]],
) -> None:
    """Test the full connect CLI path: argument parsing -> agent resolution -> tmux attach."""
    agent_name = f"test-connect-cli-tmux-{int(time.time())}"
    session_name = f"{mngr_test_prefix}{agent_name}"

    with tmux_session_cleanup(session_name):
        create_test_agent_via_cli(
            cli_runner, temp_work_dir, mngr_test_prefix, plugin_manager, agent_name, agent_cmd="sleep 493827"
        )

        result = cli_runner.invoke(
            connect,
            [agent_name],
            obj=plugin_manager,
            catch_exceptions=False,
        )
        assert result.exit_code == 0, f"Connect failed with output: {result.output}"

        # Verify the CLI resolved the agent and called tmux attach with the right session
        assert len(intercepted_execvp_calls) == 1
        assert intercepted_execvp_calls[0] == ("tmux", ["tmux", "attach", "-t", session_name])


def test_connect_via_cli_group(
    cli_runner: CliRunner,
    temp_work_dir: Path,
    mngr_test_prefix: str,
    plugin_manager: pluggy.PluginManager,
    intercepted_execvp_calls: list[tuple[str, list[str]]],
) -> None:
    """Test calling connect through the main CLI group."""
    agent_name = f"test-connect-cli-{int(time.time())}"
    session_name = f"{mngr_test_prefix}{agent_name}"

    try:
        # First create an agent
        create_result = cli_runner.invoke(
            cli,
            [
                "create",
                "--name",
                agent_name,
                "--agent-cmd",
                "sleep 918273",
                "--source",
                str(temp_work_dir),
                "--no-connect",
                "--await-ready",
                "--no-copy-work-dir",
                "--no-ensure-clean",
            ],
            obj=plugin_manager,
            catch_exceptions=False,
        )
        assert create_result.exit_code == 0, f"Create failed with: {create_result.output}"

        cli_runner.invoke(
            cli,
            ["connect", agent_name],
            obj=plugin_manager,
            catch_exceptions=False,
        )

        assert len(intercepted_execvp_calls) == 1
        assert intercepted_execvp_calls[0] == ("tmux", ["tmux", "attach", "-t", session_name])

    finally:
        cleanup_tmux_session(session_name)


def test_connect_start_restarts_stopped_agent(
    cli_runner: CliRunner,
    temp_work_dir: Path,
    mngr_test_prefix: str,
    plugin_manager: pluggy.PluginManager,
    intercepted_execvp_calls: list[tuple[str, list[str]]],
) -> None:
    """Test that --start (default) automatically restarts a stopped agent via the CLI."""
    agent_name = f"test-connect-start-{int(time.time())}"
    session_name = f"{mngr_test_prefix}{agent_name}"

    try:
        # Create an agent
        create_result = cli_runner.invoke(
            create,
            [
                "--name",
                agent_name,
                "--agent-cmd",
                "sleep 736291",
                "--source",
                str(temp_work_dir),
                "--no-connect",
                "--await-ready",
                "--no-copy-work-dir",
                "--no-ensure-clean",
            ],
            obj=plugin_manager,
            catch_exceptions=False,
        )
        assert create_result.exit_code == 0, f"Create failed with: {create_result.output}"
        assert tmux_session_exists(session_name), f"Expected tmux session {session_name} to exist"

        # Kill the tmux session to simulate a stopped agent
        cleanup_tmux_session(session_name)
        assert not tmux_session_exists(session_name), f"Expected tmux session {session_name} to be killed"

        # Connect with --start (default), which should restart the agent
        cli_runner.invoke(
            connect,
            [agent_name],
            obj=plugin_manager,
            catch_exceptions=False,
        )

        # Verify the tmux session was recreated before attaching
        assert tmux_session_exists(session_name), f"Expected tmux session {session_name} to be restarted"
        assert len(intercepted_execvp_calls) == 1
        assert intercepted_execvp_calls[0] == ("tmux", ["tmux", "attach", "-t", session_name])

    finally:
        cleanup_tmux_session(session_name)


def test_connect_no_start_raises_error_for_stopped_agent(
    cli_runner: CliRunner,
    temp_work_dir: Path,
    mngr_test_prefix: str,
    plugin_manager: pluggy.PluginManager,
) -> None:
    """Test that --no-start raises UserInputError when agent is stopped."""
    agent_name = f"test-connect-nostart-{int(time.time())}"
    session_name = f"{mngr_test_prefix}{agent_name}"

    try:
        # Create an agent
        create_result = cli_runner.invoke(
            create,
            [
                "--name",
                agent_name,
                "--agent-cmd",
                "sleep 847362",
                "--source",
                str(temp_work_dir),
                "--no-connect",
                "--await-ready",
                "--no-copy-work-dir",
                "--no-ensure-clean",
            ],
            obj=plugin_manager,
            catch_exceptions=False,
        )
        assert create_result.exit_code == 0, f"Create failed with: {create_result.output}"

        # Kill the tmux session to simulate a stopped agent
        cleanup_tmux_session(session_name)
        assert not tmux_session_exists(session_name), f"Expected tmux session {session_name} to be killed"

        # Connect with --no-start, which should raise an error
        result = cli_runner.invoke(
            connect,
            [agent_name, "--no-start"],
            obj=plugin_manager,
        )

        assert result.exit_code != 0
        assert result.exception is not None
        assert "stopped" in str(result.output).lower()
        assert "automatic starting is disabled" in str(result.output)

    finally:
        cleanup_tmux_session(session_name)


def test_connect_cli_non_interactive_selects_most_recent_agent(
    cli_runner: CliRunner,
    temp_work_dir: Path,
    mngr_test_prefix: str,
    plugin_manager: pluggy.PluginManager,
    intercepted_execvp_calls: list[tuple[str, list[str]]],
) -> None:
    """Test the non-interactive code path selects the most recently created agent.

    When stdin is not a tty (simulated by providing input=""), the connect
    command should detect non-interactive mode, call list_agents, sort by
    create_time descending, and select the most recently created agent.
    """
    agent_name_old = f"test-connect-old-{int(time.time())}"
    agent_name_new = f"test-connect-new-{int(time.time())}"
    session_old = f"{mngr_test_prefix}{agent_name_old}"
    session_new = f"{mngr_test_prefix}{agent_name_new}"

    with tmux_session_cleanup(session_old), tmux_session_cleanup(session_new):
        create_test_agent_via_cli(
            cli_runner, temp_work_dir, mngr_test_prefix, plugin_manager, agent_name_old, agent_cmd="sleep 192837"
        )
        create_test_agent_via_cli(
            cli_runner, temp_work_dir, mngr_test_prefix, plugin_manager, agent_name_new, agent_cmd="sleep 283746"
        )

        cli_runner.invoke(
            connect,
            [],
            obj=plugin_manager,
            catch_exceptions=False,
            # Providing input="" makes stdin non-tty, triggering the non-interactive path
            input="",
        )

        # Verify the CLI selected the most recently created agent
        assert len(intercepted_execvp_calls) == 1
        assert intercepted_execvp_calls[0] == ("tmux", ["tmux", "attach", "-t", session_new])


# =============================================================================
# Tests for _build_connection_options (CLI option to ConnectionOptions mapping)
# =============================================================================


def test_build_connection_options_allow_unknown_host_true(
    default_connect_cli_opts: ConnectCliOptions,
) -> None:
    """Test that allow_unknown_host=True produces is_unknown_host_allowed=True."""
    opts = default_connect_cli_opts.model_copy_update(
        to_update(default_connect_cli_opts.field_ref().allow_unknown_host, True),
    )

    connection_opts = _build_connection_options(opts)

    assert connection_opts.is_unknown_host_allowed is True


def test_build_connection_options_allow_unknown_host_default(
    default_connect_cli_opts: ConnectCliOptions,
) -> None:
    """Test that is_unknown_host_allowed defaults to False."""
    connection_opts = _build_connection_options(default_connect_cli_opts)

    assert connection_opts.is_unknown_host_allowed is False


def test_build_connection_options_maps_all_fields(
    default_connect_cli_opts: ConnectCliOptions,
) -> None:
    """Test that all ConnectCliOptions fields are correctly mapped to ConnectionOptions."""
    opts = default_connect_cli_opts.model_copy_update(
        to_update(default_connect_cli_opts.field_ref().reconnect, False),
        to_update(default_connect_cli_opts.field_ref().retry, 5),
        to_update(default_connect_cli_opts.field_ref().retry_delay, "10s"),
        to_update(default_connect_cli_opts.field_ref().attach_command, "bash"),
        to_update(default_connect_cli_opts.field_ref().allow_unknown_host, True),
    )

    connection_opts = _build_connection_options(opts)

    assert connection_opts.is_reconnect is False
    assert connection_opts.retry_count == 5
    assert connection_opts.retry_delay == "10s"
    assert connection_opts.attach_command == "bash"
    assert connection_opts.is_unknown_host_allowed is True


# =============================================================================
# Unit tests for filter_agents
# =============================================================================


def test_filter_agents_no_filters() -> None:
    """Test filter_agents with no filters applied."""
    agents = [
        make_test_agent_info("alpha", AgentLifecycleState.RUNNING),
        make_test_agent_info("beta", AgentLifecycleState.STOPPED),
        make_test_agent_info("gamma", AgentLifecycleState.RUNNING),
    ]

    result = filter_agents(agents, hide_stopped=False, search_query="")
    assert len(result) == 3


def test_filter_agents_hide_stopped() -> None:
    """Test filter_agents with hide_stopped=True."""
    agents = [
        make_test_agent_info("alpha", AgentLifecycleState.RUNNING),
        make_test_agent_info("beta", AgentLifecycleState.STOPPED),
        make_test_agent_info("gamma", AgentLifecycleState.RUNNING),
    ]

    result = filter_agents(agents, hide_stopped=True, search_query="")
    assert len(result) == 2
    assert all(a.state != AgentLifecycleState.STOPPED for a in result)


def test_filter_agents_search_query() -> None:
    """Test filter_agents with search query."""
    agents = [
        make_test_agent_info("test-alpha", AgentLifecycleState.RUNNING),
        make_test_agent_info("test-beta", AgentLifecycleState.RUNNING),
        make_test_agent_info("other", AgentLifecycleState.RUNNING),
    ]

    # Case-insensitive search
    result = filter_agents(agents, hide_stopped=False, search_query="ALPHA")
    assert len(result) == 1
    assert result[0].name == AgentName("test-alpha")

    # Partial match
    result = filter_agents(agents, hide_stopped=False, search_query="test")
    assert len(result) == 2


def test_filter_agents_combined_filters() -> None:
    """Test filter_agents with both filters applied."""
    agents = [
        make_test_agent_info("test-alpha", AgentLifecycleState.RUNNING),
        make_test_agent_info("test-beta", AgentLifecycleState.STOPPED),
        make_test_agent_info("other", AgentLifecycleState.RUNNING),
    ]

    result = filter_agents(agents, hide_stopped=True, search_query="test")
    assert len(result) == 1
    assert result[0].name == AgentName("test-alpha")


# =============================================================================
# Unit tests for build_status_text
# =============================================================================


def test_build_status_text_default() -> None:
    """Test build_status_text with default state."""
    status = build_status_text(search_query="", hide_stopped=False)
    assert "Status: Ready" in status
    assert "Type to search" in status
    assert "Filter: All agents" in status


def test_build_status_text_with_search() -> None:
    """Test build_status_text with search query."""
    status = build_status_text(search_query="test", hide_stopped=False)
    assert "Search: test" in status
    assert "Type to search" not in status


def test_build_status_text_hide_stopped() -> None:
    """Test build_status_text with hide_stopped=True."""
    status = build_status_text(search_query="", hide_stopped=True)
    assert "Filter: Hiding stopped" in status


# =============================================================================
# Unit tests for handle_search_key
# =============================================================================


def test_handle_search_key_backspace() -> None:
    """Test handle_search_key removes last character on backspace."""
    new_query, should_refresh = handle_search_key(
        key="backspace",
        is_printable=False,
        character=None,
        current_query="test",
    )
    assert new_query == "tes"
    assert should_refresh is True


def test_handle_search_key_backspace_empty() -> None:
    """Test handle_search_key does nothing on backspace with empty query."""
    new_query, should_refresh = handle_search_key(
        key="backspace",
        is_printable=False,
        character=None,
        current_query="",
    )
    assert new_query == ""
    assert should_refresh is False


def test_handle_search_key_printable() -> None:
    """Test handle_search_key adds printable characters."""
    new_query, should_refresh = handle_search_key(
        key="a",
        is_printable=True,
        character="a",
        current_query="test",
    )
    assert new_query == "testa"
    assert should_refresh is True


def test_handle_search_key_other() -> None:
    """Test handle_search_key passes through other keys."""
    new_query, should_refresh = handle_search_key(
        key="up",
        is_printable=False,
        character=None,
        current_query="test",
    )
    assert new_query == "test"
    assert should_refresh is False


# =============================================================================
# Urwid Agent Selector UI Tests
# =============================================================================


def test_create_selectable_agent_item_displays_agent_info() -> None:
    """Test that _create_selectable_agent_item creates a selectable widget."""
    agent = make_test_agent_info("test-agent", AgentLifecycleState.RUNNING)

    widget = _create_selectable_agent_item(agent, name_width=20, state_width=10)

    assert isinstance(widget, AttrMap)
    inner_widget = widget.original_widget
    # Widget is now a SelectableIcon so ListBox can navigate between items
    assert isinstance(inner_widget, SelectableIcon)
    assert widget.selectable() is True


def test_create_selectable_agent_item_stopped_state() -> None:
    """Test that _create_selectable_agent_item shows stopped state correctly."""
    agent = make_test_agent_info("stopped-agent", AgentLifecycleState.STOPPED)

    widget = _create_selectable_agent_item(agent, name_width=20, state_width=10)

    inner_widget = widget.original_widget
    assert isinstance(inner_widget, SelectableIcon)
    assert widget.selectable() is True


def _create_test_selector_state(agents: list[AgentInfo]) -> AgentSelectorState:
    """Create an AgentSelectorState for testing without refreshing the list."""
    list_walker: SimpleFocusListWalker[AttrMap] = SimpleFocusListWalker([])
    status_text = Text("")
    # Calculate reasonable default widths for tests
    name_width = max((len(str(a.name)) for a in agents), default=10)
    state_width = max((len(a.state.value) for a in agents), default=7)
    return AgentSelectorState(
        agents=agents,
        list_walker=list_walker,
        status_text=status_text,
        name_width=name_width,
        state_width=state_width,
    )


def _create_and_refresh_test_state(agents: list[AgentInfo]) -> AgentSelectorState:
    """Create an AgentSelectorState and refresh it for testing."""
    state = _create_test_selector_state(agents)
    _refresh_agent_list(state)
    return state


def test_agent_selector_state_initializes_with_defaults() -> None:
    """Test that AgentSelectorState initializes with correct defaults."""
    agents = [
        make_test_agent_info("alpha", AgentLifecycleState.RUNNING),
        make_test_agent_info("beta", AgentLifecycleState.STOPPED),
    ]

    state = _create_test_selector_state(agents)

    assert state.agents == agents
    assert state.filtered_agents == []
    assert state.result is None
    assert state.hide_stopped is False
    assert state.search_query == ""


def test_refresh_agent_list_populates_list_walker() -> None:
    """Test that _refresh_agent_list populates the list walker with agents."""
    agents = [
        make_test_agent_info("alpha", AgentLifecycleState.RUNNING),
        make_test_agent_info("beta", AgentLifecycleState.RUNNING),
        make_test_agent_info("gamma", AgentLifecycleState.STOPPED),
    ]
    state = _create_and_refresh_test_state(agents)

    assert len(state.list_walker) == 3
    assert len(state.filtered_agents) == 3


def test_refresh_agent_list_applies_hide_stopped_filter() -> None:
    """Test that _refresh_agent_list respects hide_stopped setting."""
    agents = [
        make_test_agent_info("alpha", AgentLifecycleState.RUNNING),
        make_test_agent_info("beta", AgentLifecycleState.STOPPED),
    ]
    state = _create_test_selector_state(agents)
    state.hide_stopped = True

    _refresh_agent_list(state)

    assert len(state.list_walker) == 1
    assert len(state.filtered_agents) == 1
    assert state.filtered_agents[0].name == AgentName("alpha")


def test_refresh_agent_list_applies_search_filter() -> None:
    """Test that _refresh_agent_list respects search_query setting."""
    agents = [
        make_test_agent_info("alpha-test", AgentLifecycleState.RUNNING),
        make_test_agent_info("beta-prod", AgentLifecycleState.RUNNING),
    ]
    state = _create_test_selector_state(agents)
    state.search_query = "alpha"

    _refresh_agent_list(state)

    assert len(state.list_walker) == 1
    assert len(state.filtered_agents) == 1
    assert state.filtered_agents[0].name == AgentName("alpha-test")


def test_refresh_agent_list_updates_status_text() -> None:
    """Test that _refresh_agent_list updates the status text widget."""
    agents = [make_test_agent_info("alpha", AgentLifecycleState.RUNNING)]
    state = _create_test_selector_state(agents)
    state.search_query = "search-term"

    _refresh_agent_list(state)

    status_content = state.status_text.get_text()[0]
    assert "search-term" in status_content


def test_refresh_agent_list_sets_focus_on_first_item() -> None:
    """Test that _refresh_agent_list sets focus on the first item when items exist."""
    agents = [
        make_test_agent_info("alpha", AgentLifecycleState.RUNNING),
        make_test_agent_info("beta", AgentLifecycleState.RUNNING),
    ]
    state = _create_and_refresh_test_state(agents)

    _, focus_index = state.list_walker.get_focus()
    assert focus_index == 0


def test_handle_selector_input_ctrl_r_toggles_hide_stopped() -> None:
    """Test that Ctrl+R toggles the hide_stopped filter."""
    agents = [
        make_test_agent_info("alpha", AgentLifecycleState.RUNNING),
        make_test_agent_info("beta", AgentLifecycleState.STOPPED),
    ]
    state = _create_and_refresh_test_state(agents)

    assert state.hide_stopped is False
    assert len(state.filtered_agents) == 2

    _handle_selector_input(state, "ctrl r")

    assert state.hide_stopped is True
    assert len(state.filtered_agents) == 1


def test_handle_selector_input_ctrl_c_clears_search_query() -> None:
    """Test that Ctrl+C clears the search query when not empty."""
    agents = [make_test_agent_info("alpha", AgentLifecycleState.RUNNING)]
    state = _create_and_refresh_test_state(agents)
    state.search_query = "test"

    _handle_selector_input(state, "ctrl c")

    assert state.search_query == ""


def test_handle_selector_input_ctrl_c_double_tap_exits() -> None:
    """Test that double Ctrl+C (within 500ms) raises ExitMainLoop."""
    agents = [make_test_agent_info("alpha", AgentLifecycleState.RUNNING)]
    state = _create_and_refresh_test_state(agents)

    # First Ctrl-c with empty query records the time
    _handle_selector_input(state, "ctrl c")

    # Second Ctrl-c within 500ms should exit
    with pytest.raises(ExitMainLoop):
        _handle_selector_input(state, "ctrl c")


def test_handle_selector_input_enter_selects_focused_agent() -> None:
    """Test that Enter selects the currently focused agent."""
    agents = [
        make_test_agent_info("alpha", AgentLifecycleState.RUNNING),
        make_test_agent_info("beta", AgentLifecycleState.RUNNING),
    ]
    state = _create_and_refresh_test_state(agents)

    with pytest.raises(ExitMainLoop):
        _handle_selector_input(state, "enter")

    assert state.result is not None
    assert state.result.name == AgentName("alpha")


def test_handle_selector_input_enter_with_empty_list_sets_no_result() -> None:
    """Test that Enter with empty list doesn't set a result."""
    agents: list[AgentInfo] = []
    state = _create_and_refresh_test_state(agents)

    with pytest.raises(ExitMainLoop):
        _handle_selector_input(state, "enter")

    assert state.result is None


def test_handle_selector_input_arrow_keys_pass_through() -> None:
    """Test that arrow keys pass through (return False) for ListBox to handle."""
    agents = [make_test_agent_info("alpha", AgentLifecycleState.RUNNING)]
    state = _create_and_refresh_test_state(agents)

    # Arrow keys should return False to let ListBox handle them
    assert _handle_selector_input(state, "up") is False
    assert _handle_selector_input(state, "down") is False
    assert _handle_selector_input(state, "page up") is False
    assert _handle_selector_input(state, "page down") is False


def test_handle_selector_input_printable_key_updates_search() -> None:
    """Test that printable keys update the search query."""
    agents = [
        make_test_agent_info("xyz-agent", AgentLifecycleState.RUNNING),
        make_test_agent_info("foo-agent", AgentLifecycleState.RUNNING),
    ]
    state = _create_and_refresh_test_state(agents)

    _handle_selector_input(state, "x")

    assert state.search_query == "x"
    assert len(state.filtered_agents) == 1
    assert state.filtered_agents[0].name == AgentName("xyz-agent")


def test_handle_selector_input_backspace_removes_last_character() -> None:
    """Test that backspace removes the last character from search query."""
    agents = [make_test_agent_info("alpha", AgentLifecycleState.RUNNING)]
    state = _create_and_refresh_test_state(agents)
    state.search_query = "alph"

    _handle_selector_input(state, "backspace")

    assert state.search_query == "alp"


def test_selector_input_handler_calls_handle_selector_input() -> None:
    """Test that SelectorInputHandler delegates to _handle_selector_input."""
    agents = [make_test_agent_info("alpha", AgentLifecycleState.RUNNING)]
    state = _create_and_refresh_test_state(agents)
    handler = SelectorInputHandler(state=state)

    handler("a")

    assert state.search_query == "a"


def test_selector_input_handler_ignores_mouse_events() -> None:
    """Test that SelectorInputHandler returns None for mouse events (tuples)."""
    agents = [make_test_agent_info("alpha", AgentLifecycleState.RUNNING)]
    state = _create_and_refresh_test_state(agents)
    handler = SelectorInputHandler(state=state)

    result = handler(("mouse press", 1, 0, 0))

    assert result is None
    assert state.search_query == ""


def test_select_agent_interactively_returns_none_for_empty_list() -> None:
    """Test that select_agent_interactively returns None when given empty list."""
    agents: list[AgentInfo] = []

    result = select_agent_interactively(agents)

    assert result is None



from pathlib import Path

from imbue.mngr.api.create import CreateAgentOptions
from imbue.mngr.api.message import MessageResult
from imbue.mngr.api.message import _agent_to_cel_context
from imbue.mngr.api.message import send_message_to_agents
from imbue.mngr.config.data_types import MngrContext
from imbue.mngr.hosts.host import Host
from imbue.mngr.primitives import AgentLifecycleState
from imbue.mngr.primitives import AgentName
from imbue.mngr.primitives import AgentTypeName
from imbue.mngr.primitives import CommandString
from imbue.mngr.primitives import ErrorBehavior
from imbue.mngr.primitives import HostName
from imbue.mngr.providers.local.instance import LocalProviderInstance


def test_message_result_initializes_with_empty_lists() -> None:
    """Test that MessageResult initializes with empty lists."""
    result = MessageResult()
    assert result.successful_agents == []
    assert result.failed_agents == []


def test_message_result_can_add_successful_agent() -> None:
    """Test that we can add successful agents to the result."""
    result = MessageResult()
    result.successful_agents.append("test-agent")
    assert result.successful_agents == ["test-agent"]


def test_message_result_can_add_failed_agent() -> None:
    """Test that we can add failed agents to the result."""
    result = MessageResult()
    result.failed_agents.append(("test-agent", "error message"))
    assert result.failed_agents == [("test-agent", "error message")]


def test_agent_to_cel_context_returns_expected_fields(
    temp_work_dir: Path,
    local_provider: LocalProviderInstance,
) -> None:
    """Test that _agent_to_cel_context returns the expected fields."""
    host = local_provider.create_host(HostName("localhost"))
    assert isinstance(host, Host)

    agent = host.create_agent_state(
        work_dir_path=temp_work_dir,
        options=CreateAgentOptions(
            name=AgentName("cel-test-agent"),
            agent_type=AgentTypeName("generic"),
            command=CommandString("sleep 1"),
        ),
    )

    context = _agent_to_cel_context(agent, "local")

    assert context["id"] == str(agent.id)
    assert context["name"] == "cel-test-agent"
    assert context["type"] == "generic"
    assert context["state"] == AgentLifecycleState.STOPPED.value
    assert context["host"]["provider"] == "local"
    assert context["host"]["id"] == str(agent.host_id)


def test_send_message_to_agents_returns_empty_result_when_no_agents_match(
    temp_mngr_ctx: MngrContext,
) -> None:
    """Test that send_message returns empty result when no agents match filters."""
    result = send_message_to_agents(
        mngr_ctx=temp_mngr_ctx,
        message_content="Hello",
        include_filters=('name == "nonexistent-agent"',),
        all_agents=False,
    )

    assert result.successful_agents == []
    assert result.failed_agents == []


def test_send_message_to_agents_calls_success_callback(
    temp_work_dir: Path,
    temp_mngr_ctx: MngrContext,
    local_provider: LocalProviderInstance,
) -> None:
    """Test that send_message calls the success callback when message is sent."""
    host = local_provider.create_host(HostName("localhost"))
    assert isinstance(host, Host)

    agent = host.create_agent_state(
        work_dir_path=temp_work_dir,
        options=CreateAgentOptions(
            name=AgentName("message-test"),
            agent_type=AgentTypeName("generic"),
            command=CommandString("sleep 847264"),
        ),
    )

    # Start the agent
    host.start_agents([agent.id])

    success_agents: list[str] = []
    error_agents: list[tuple[str, str]] = []

    result = send_message_to_agents(
        mngr_ctx=temp_mngr_ctx,
        message_content="Hello from test",
        all_agents=True,
        on_success=lambda name: success_agents.append(name),
        on_error=lambda name, err: error_agents.append((name, err)),
    )

    # Clean up
    host.destroy_agent(agent)

    assert "message-test" in result.successful_agents
    assert "message-test" in success_agents


def test_send_message_to_agents_fails_for_stopped_agent(
    temp_work_dir: Path,
    temp_mngr_ctx: MngrContext,
    local_provider: LocalProviderInstance,
) -> None:
    """Test that sending message to stopped agent fails."""
    host = local_provider.create_host(HostName("localhost"))
    assert isinstance(host, Host)

    agent = host.create_agent_state(
        work_dir_path=temp_work_dir,
        options=CreateAgentOptions(
            name=AgentName("stopped-test"),
            agent_type=AgentTypeName("generic"),
            command=CommandString("sleep 847265"),
        ),
    )

    # Don't start the agent - it should be stopped

    result = send_message_to_agents(
        mngr_ctx=temp_mngr_ctx,
        message_content="Hello",
        all_agents=True,
        error_behavior=ErrorBehavior.CONTINUE,
    )

    # Clean up
    host.destroy_agent(agent)

    # Should have failed because agent has no tmux session
    assert len(result.failed_agents) == 1
    assert result.failed_agents[0][0] == "stopped-test"
    assert "no tmux session" in result.failed_agents[0][1]


def test_send_message_to_agents_starts_stopped_agent_when_start_desired(
    temp_work_dir: Path,
    temp_mngr_ctx: MngrContext,
    local_provider: LocalProviderInstance,
) -> None:
    """Test that send_message auto-starts a stopped agent when is_start_desired=True."""
    host = local_provider.create_host(HostName("localhost"))
    assert isinstance(host, Host)

    agent = host.create_agent_state(
        work_dir_path=temp_work_dir,
        options=CreateAgentOptions(
            name=AgentName("start-test"),
            agent_type=AgentTypeName("generic"),
            command=CommandString("sleep 847268"),
        ),
    )

    # Don't start the agent - it should be stopped
    assert agent.get_lifecycle_state() == AgentLifecycleState.STOPPED

    success_agents: list[str] = []
    error_agents: list[tuple[str, str]] = []

    result = send_message_to_agents(
        mngr_ctx=temp_mngr_ctx,
        message_content="Hello with auto-start",
        all_agents=True,
        is_start_desired=True,
        on_success=lambda name: success_agents.append(name),
        on_error=lambda name, err: error_agents.append((name, err)),
    )

    # Clean up
    host.destroy_agent(agent)

    # Agent should have been started and message sent successfully
    assert "start-test" in result.successful_agents
    assert "start-test" in success_agents
    assert len(error_agents) == 0


def test_send_message_to_agents_with_include_filter(
    temp_work_dir: Path,
    temp_mngr_ctx: MngrContext,
    local_provider: LocalProviderInstance,
) -> None:
    """Test that send_message respects include filters."""
    host = local_provider.create_host(HostName("localhost"))
    assert isinstance(host, Host)

    # Create two agents
    agent1 = host.create_agent_state(
        work_dir_path=temp_work_dir,
        options=CreateAgentOptions(
            name=AgentName("filter-test-1"),
            agent_type=AgentTypeName("generic"),
            command=CommandString("sleep 847266"),
        ),
    )
    agent2 = host.create_agent_state(
        work_dir_path=temp_work_dir,
        options=CreateAgentOptions(
            name=AgentName("filter-test-2"),
            agent_type=AgentTypeName("generic"),
            command=CommandString("sleep 847267"),
        ),
    )

    # Start both agents
    host.start_agents([agent1.id, agent2.id])

    # Send message only to agent1 using filter
    result = send_message_to_agents(
        mngr_ctx=temp_mngr_ctx,
        message_content="Hello filtered",
        include_filters=('name == "filter-test-1"',),
        all_agents=False,
    )

    # Clean up
    host.destroy_agent(agent1)
    host.destroy_agent(agent2)

    # Only agent1 should have received the message
    assert "filter-test-1" in result.successful_agents
    assert "filter-test-2" not in result.successful_agents



"""Unit tests for gc API functions."""

from datetime import datetime
from datetime import timedelta
from datetime import timezone
from pathlib import Path

import pytest

from imbue.mngr.api.data_types import GcResult
from imbue.mngr.api.gc import _apply_cel_filters
from imbue.mngr.api.gc import _resource_to_cel_context
from imbue.mngr.api.gc import gc_machines
from imbue.mngr.config.data_types import MngrContext
from imbue.mngr.hosts.offline_host import OfflineHost
from imbue.mngr.interfaces.data_types import CertifiedHostData
from imbue.mngr.interfaces.data_types import LogFileInfo
from imbue.mngr.interfaces.data_types import SizeBytes
from imbue.mngr.interfaces.data_types import SnapshotInfo
from imbue.mngr.interfaces.data_types import VolumeInfo
from imbue.mngr.primitives import AgentId
from imbue.mngr.primitives import ErrorBehavior
from imbue.mngr.primitives import HostId
from imbue.mngr.primitives import HostState
from imbue.mngr.primitives import ProviderInstanceName
from imbue.mngr.primitives import SnapshotId
from imbue.mngr.primitives import SnapshotName
from imbue.mngr.primitives import VolumeId
from imbue.mngr.providers.local.instance import LocalProviderInstance
from imbue.mngr.providers.mock_provider_test import MockProviderInstance
from imbue.mngr.providers.mock_provider_test import make_offline_host
from imbue.mngr.utils.cel_utils import compile_cel_filters


def test_resource_to_cel_context_for_path(tmp_path: Path) -> None:
    """Test converting LogFileInfo objects to CEL context."""
    test_file = tmp_path / "test.txt"
    test_file.write_text("test content")

    created_at = datetime.fromtimestamp(test_file.stat().st_ctime, tz=timezone.utc)
    log_file_info = LogFileInfo(path=test_file, size_bytes=SizeBytes(len("test content")), created_at=created_at)
    context = _resource_to_cel_context(log_file_info)

    assert context["type"] == "logfile"
    assert context["name"] == "test.txt"
    assert context["path"] == str(test_file)
    assert context["size"] == len("test content")
    assert "age" in context
    assert isinstance(context["age"], (int, float))


def test_resource_to_cel_context_for_snapshot() -> None:
    """Test converting SnapshotInfo to CEL context."""
    created_time = datetime.now(timezone.utc) - timedelta(days=7)
    snapshot = SnapshotInfo(
        id=SnapshotId("snap-test"),
        name=SnapshotName("test-snapshot"),
        created_at=created_time,
        size_bytes=1000000,
    )

    context = _resource_to_cel_context(snapshot)

    assert context["type"] == "snapshot"
    assert context["name"] == "test-snapshot"
    assert context["size"] == 1000000
    assert context["size_bytes"] == 1000000
    assert "age" in context
    assert 7 * 24 * 3600 - 10 < context["age"] < 7 * 24 * 3600 + 10


def test_resource_to_cel_context_for_volume() -> None:
    """Test converting VolumeInfo to CEL context."""
    created_time = datetime.now(timezone.utc) - timedelta(hours=2)
    volume = VolumeInfo(
        volume_id=VolumeId.generate(),
        name="test-volume",
        size_bytes=500000000,
        created_at=created_time,
        host_id=HostId.generate(),
    )

    context = _resource_to_cel_context(volume)

    assert context["type"] == "volume"
    assert context["name"] == "test-volume"
    assert context["size"] == 500000000
    assert context["size_bytes"] == 500000000
    assert "age" in context
    assert 2 * 3600 - 10 < context["age"] < 2 * 3600 + 10


def test_compile_and_apply_cel_filters_include() -> None:
    """Test CEL filter compilation and application with include filters."""
    snapshot = SnapshotInfo(
        id=SnapshotId("snap-test"),
        name=SnapshotName("large-snapshot"),
        created_at=datetime.now(timezone.utc) - timedelta(days=10),
        size_bytes=2000000000,
    )

    include_filters, exclude_filters = compile_cel_filters(
        include_filters=["size > 1000000000"],
        exclude_filters=[],
    )

    assert _apply_cel_filters(snapshot, include_filters, exclude_filters)


def test_compile_and_apply_cel_filters_exclude() -> None:
    """Test CEL filter compilation and application with exclude filters."""
    snapshot = SnapshotInfo(
        id=SnapshotId("snap-test"),
        name=SnapshotName("old-snapshot"),
        created_at=datetime.now(timezone.utc) - timedelta(days=30),
        size_bytes=500000,
    )

    include_filters, exclude_filters = compile_cel_filters(
        include_filters=[],
        exclude_filters=["age > 2592000"],
    )

    assert not _apply_cel_filters(snapshot, include_filters, exclude_filters)


def test_compile_and_apply_cel_filters_name_matching() -> None:
    """Test CEL filter with name pattern matching."""
    snapshot = SnapshotInfo(
        id=SnapshotId("snap-test"),
        name=SnapshotName("temp-snapshot-123"),
        created_at=datetime.now(timezone.utc),
        size_bytes=100000,
    )

    include_filters, exclude_filters = compile_cel_filters(
        include_filters=[],
        exclude_filters=["name.startsWith('temp')"],
    )

    assert not _apply_cel_filters(snapshot, include_filters, exclude_filters)


def test_compile_and_apply_cel_filters_multiple() -> None:
    """Test CEL filter with multiple conditions."""
    snapshot = SnapshotInfo(
        id=SnapshotId("snap-test"),
        name=SnapshotName("prod-snapshot"),
        created_at=datetime.now(timezone.utc) - timedelta(days=15),
        size_bytes=3000000000,
    )

    include_filters, exclude_filters = compile_cel_filters(
        include_filters=["size > 1000000000", "age > 604800"],
        exclude_filters=["name.startsWith('temp')"],
    )

    assert _apply_cel_filters(snapshot, include_filters, exclude_filters)


def test_compile_and_apply_cel_filters_empty_filters() -> None:
    """Test CEL filter with no filters returns True."""
    snapshot = SnapshotInfo(
        id=SnapshotId("snap-test"),
        name=SnapshotName("any-snapshot"),
        created_at=datetime.now(timezone.utc),
        size_bytes=1000,
    )

    include_filters, exclude_filters = compile_cel_filters(
        include_filters=[],
        exclude_filters=[],
    )

    assert _apply_cel_filters(snapshot, include_filters, exclude_filters)


def test_compile_and_apply_cel_filters_include_not_matching() -> None:
    """Test CEL filter with include that doesn't match returns False."""
    snapshot = SnapshotInfo(
        id=SnapshotId("snap-test"),
        name=SnapshotName("small-snapshot"),
        created_at=datetime.now(timezone.utc),
        size_bytes=100,
    )

    include_filters, exclude_filters = compile_cel_filters(
        include_filters=["size > 1000000000"],
        exclude_filters=[],
    )

    assert not _apply_cel_filters(snapshot, include_filters, exclude_filters)


def test_gc_machines_skips_local_hosts(local_provider: LocalProviderInstance, temp_mngr_ctx: MngrContext) -> None:
    """Test that gc_machines skips local hosts even when they have no agents."""
    result = GcResult()

    gc_machines(
        mngr_ctx=temp_mngr_ctx,
        providers=[local_provider],
        include_filters=(),
        exclude_filters=(),
        dry_run=False,
        error_behavior=ErrorBehavior.ABORT,
        result=result,
    )

    # Local host should be skipped, not destroyed
    assert len(result.machines_destroyed) == 0
    assert len(result.errors) == 0


# =========================================================================
# gc_machines offline host deletion tests
# =========================================================================


@pytest.fixture
def gc_mock_provider(temp_host_dir: Path, temp_mngr_ctx: MngrContext) -> MockProviderInstance:
    """Create a MockProviderInstance for gc_machines tests."""
    return MockProviderInstance(
        name=ProviderInstanceName("test-provider"),
        host_dir=temp_host_dir,
        mngr_ctx=temp_mngr_ctx,
    )


def _make_offline_host(
    provider: MockProviderInstance,
    mngr_ctx: MngrContext,
    *,
    days_old: int = 14,
    stop_reason: str | None = HostState.STOPPED.value,
    failure_reason: str | None = None,
) -> OfflineHost:
    """Create an offline host with configurable age and state."""
    stopped_at = datetime.now(timezone.utc) - timedelta(days=days_old)
    certified_data = CertifiedHostData(
        host_id=str(HostId.generate()),
        host_name="test-host",
        stop_reason=stop_reason,
        failure_reason=failure_reason,
        created_at=stopped_at - timedelta(hours=1),
        updated_at=stopped_at,
    )
    return make_offline_host(certified_data, provider, mngr_ctx)


def _run_gc_machines(provider: MockProviderInstance, *, dry_run: bool = False) -> GcResult:
    """Run gc_machines on a single provider and return the result."""
    result = GcResult()
    gc_machines(
        mngr_ctx=provider.mngr_ctx,
        providers=[provider],
        include_filters=(),
        exclude_filters=(),
        dry_run=dry_run,
        error_behavior=ErrorBehavior.ABORT,
        result=result,
    )
    return result


def test_gc_machines_deletes_old_offline_host_with_no_agents(
    gc_mock_provider: MockProviderInstance, temp_mngr_ctx: MngrContext
) -> None:
    """Old offline hosts with no agents are deleted to prevent data accumulation."""
    host = _make_offline_host(gc_mock_provider, temp_mngr_ctx, days_old=14)
    gc_mock_provider.mock_hosts = [host]

    result = _run_gc_machines(gc_mock_provider)

    assert len(result.machines_deleted) == 1
    assert result.machines_deleted[0].id == host.id
    assert gc_mock_provider.deleted_hosts == [host.id]


def test_gc_machines_skips_recent_offline_host(
    gc_mock_provider: MockProviderInstance, temp_mngr_ctx: MngrContext
) -> None:
    """Offline hosts stopped less than the max persisted seconds ago are not deleted."""
    host = _make_offline_host(gc_mock_provider, temp_mngr_ctx, days_old=1)
    gc_mock_provider.mock_hosts = [host]

    result = _run_gc_machines(gc_mock_provider)

    assert len(result.machines_deleted) == 0
    assert gc_mock_provider.deleted_hosts == []


def _add_mock_agent(provider: MockProviderInstance) -> None:
    """Add a mock agent to the provider so hosts appear to have agents."""
    agent_id = AgentId.generate()
    provider.mock_agent_data = [{"id": str(agent_id), "name": "test-agent"}]


def test_gc_machines_deletes_old_crashed_host_with_agents(
    gc_mock_provider: MockProviderInstance, temp_mngr_ctx: MngrContext
) -> None:
    """Old offline hosts in CRASHED state are deleted even if they have agents."""
    # None stop_reason means the host CRASHED
    host = _make_offline_host(gc_mock_provider, temp_mngr_ctx, stop_reason=None)
    _add_mock_agent(gc_mock_provider)
    gc_mock_provider.mock_hosts = [host]

    result = _run_gc_machines(gc_mock_provider)

    assert len(result.machines_deleted) == 1
    assert gc_mock_provider.deleted_hosts == [host.id]


def test_gc_machines_skips_old_stopped_host_with_agents(
    gc_mock_provider: MockProviderInstance, temp_mngr_ctx: MngrContext
) -> None:
    """Old offline hosts in STOPPED state with agents are not deleted."""
    host = _make_offline_host(gc_mock_provider, temp_mngr_ctx, days_old=14)
    _add_mock_agent(gc_mock_provider)
    gc_mock_provider.mock_hosts = [host]

    result = _run_gc_machines(gc_mock_provider)

    assert len(result.machines_deleted) == 0
    assert gc_mock_provider.deleted_hosts == []


def test_gc_machines_dry_run_does_not_call_delete_host(
    gc_mock_provider: MockProviderInstance, temp_mngr_ctx: MngrContext
) -> None:
    """Dry run identifies hosts for deletion but does not actually delete them."""
    host = _make_offline_host(gc_mock_provider, temp_mngr_ctx, days_old=14)
    gc_mock_provider.mock_hosts = [host]

    result = _run_gc_machines(gc_mock_provider, dry_run=True)

    assert len(result.machines_deleted) == 1
    assert gc_mock_provider.deleted_hosts == []


def test_gc_machines_deletes_old_failed_host_with_agents(
    gc_mock_provider: MockProviderInstance, temp_mngr_ctx: MngrContext
) -> None:
    """Old offline hosts in FAILED state are deleted even if they have agents."""
    host = _make_offline_host(gc_mock_provider, temp_mngr_ctx, failure_reason="Build failed")
    _add_mock_agent(gc_mock_provider)
    gc_mock_provider.mock_hosts = [host]

    result = _run_gc_machines(gc_mock_provider)

    assert len(result.machines_deleted) == 1
    assert gc_mock_provider.deleted_hosts == [host.id]


def test_gc_machines_deletes_old_destroyed_host_with_agents(
    gc_mock_provider: MockProviderInstance, temp_mngr_ctx: MngrContext
) -> None:
    """Old offline hosts in DESTROYED state are deleted even if they have agents."""
    host = _make_offline_host(gc_mock_provider, temp_mngr_ctx)
    # Make the provider not support snapshots and not support shutdown hosts
    # so the state resolves to DESTROYED
    gc_mock_provider.mock_supports_snapshots = False
    gc_mock_provider.mock_supports_shutdown_hosts = False
    _add_mock_agent(gc_mock_provider)
    gc_mock_provider.mock_hosts = [host]

    result = _run_gc_machines(gc_mock_provider)

    assert len(result.machines_deleted) == 1
    assert gc_mock_provider.deleted_hosts == [host.id]



from pathlib import Path

from click import ClickException

from imbue.mngr.primitives import AgentId
from imbue.mngr.primitives import HostId
from imbue.mngr.primitives import HostName
from imbue.mngr.primitives import HostState
from imbue.mngr.primitives import ImageReference
from imbue.mngr.primitives import ProviderInstanceName
from imbue.mngr.primitives import SnapshotId


class BaseMngrError(Exception):
    """Base exception for all mngr errors."""


class MngrError(ClickException, BaseMngrError):
    """Base exception for all user-facing mngr errors.

    All MngrError subclasses can provide a user_help_text attribute that contains
    additional context to help the user understand and resolve the error.
    This help text is displayed by the CLI when the error is raised.
    """

    user_help_text: str | None = None

    def format_message(self) -> str:
        if self.user_help_text:
            return str(self) + "  [" + self.user_help_text + "]"
        return str(self)


class UserInputError(MngrError):
    """Raised when user input is invalid."""

    user_help_text = "Check the command syntax with 'mngr --help' or 'mngr <command> --help'."


class ParseSpecError(MngrError, ValueError):
    """Raised when parsing a specification string fails."""


class InvalidRelativePathError(MngrError, ValueError):
    """Raised when a path that should be relative is actually absolute."""

    def __init__(self, path: str) -> None:
        self.path = path
        super().__init__(f"Path must be relative, got absolute path: {path}")


class HostError(BaseMngrError):
    """Base class for host-related errors."""


class InvalidActivityTypeError(HostError, ValueError):
    """Raised when an invalid activity type is used."""


class HostConnectionError(HostError):
    """Raised when unable to connect to a host."""


class HostOfflineError(HostConnectionError):
    """Raised when unable to connect to a host because it is offline."""


class HostDataSchemaError(HostError):
    """Raised when host data.json has an incompatible schema.

    This typically happens after mngr is upgraded and the data format changed.
    """

    def __init__(self, data_path: str, validation_error: str) -> None:
        self.data_path = data_path
        self.validation_error = validation_error
        data_dir = str(Path(data_path).parent)
        message = (
            f"Host data file has incompatible schema: {data_path}\n"
            f"This usually means mngr was upgraded and the data format changed.\n"
            f"To fix, either delete the file:\n"
            f"  rm {data_path}\n"
            f"Or run:\n"
            f'  claude --add-dir {data_dir} -p "migrate {data_path} to the new schema"'
        )
        super().__init__(message)
        self.user_help_text = f"Validation error details: {validation_error}"


class CommandTimeoutError(HostError):
    """Raised when a command execution times out."""


class LockNotHeldError(HostError):
    """Raised when attempting to use a lock that is not held."""


class AgentError(BaseMngrError):
    """Base class for agent-related errors."""


class NoCommandDefinedError(AgentError, ValueError):
    """Raised when no command is defined for an agent type."""


class AgentNotFoundError(AgentError, MngrError):
    """No agent with this ID exists."""

    user_help_text = "Use 'mngr list' to see available agents."

    def __init__(self, agent_identifier: str) -> None:
        self.agent_identifier = agent_identifier
        super().__init__(f"Agent not found: {agent_identifier}")


class AgentNotFoundOnHostError(AgentError):
    """No agent with this ID exists on the specified host."""

    user_help_text = "Use 'mngr list' to see all agents and their host assignments."

    def __init__(self, agent_id: AgentId, host_id: HostId) -> None:
        self.agent_id = agent_id
        self.host_id = host_id
        super().__init__(f"Agent {agent_id} not found on host {host_id}")


class SendMessageError(AgentError):
    """Failed to send a message to an agent."""

    def __init__(self, agent_name: str, reason: str) -> None:
        self.agent_name = agent_name
        self.reason = reason
        super().__init__(f"Failed to send message to agent {agent_name}: {reason}")


class AgentStartError(AgentError):
    """Failed to start an agent's tmux session."""

    def __init__(self, agent_name: str, reason: str) -> None:
        self.agent_name = agent_name
        self.reason = reason
        super().__init__(f"Failed to start agent {agent_name}: {reason}")


class ProviderError(MngrError):
    """Base class for all provider-related errors."""


class ProviderInstanceNotFoundError(ProviderError):
    """No provider instance with this name exists."""

    user_help_text = (
        "Check your mngr configuration for available providers.\nBuilt-in providers include 'local' and 'docker'."
    )

    def __init__(self, provider_name: ProviderInstanceName) -> None:
        self.provider_name = provider_name
        super().__init__(f"Provider {provider_name} not found")


class ProviderNotAuthorizedError(ProviderError):
    """Provider instance is not authorized/authenticated."""

    def __init__(self, provider_name: ProviderInstanceName, auth_help: str | None = None) -> None:
        self.provider_name = provider_name
        message = f"Provider '{provider_name}' is not authorized."
        if auth_help:
            message = f"{message} {auth_help}"
        super().__init__(message)
        self.user_help_text = (
            f"To disable this provider, run:\n"
            f"  mngr config set --scope user providers.{provider_name}.is_enabled false\n"
            f"Or disable the provider backend entirely by removing it from enabled_backends in your config."
        )


class HostNotFoundError(ProviderError):
    """No host with this ID or name exists."""

    user_help_text = "Use 'mngr list' to see available hosts and agents."

    def __init__(self, host: HostId | HostName) -> None:
        self.host = host
        super().__init__(f"Host not found: {host}")


class HostCreationError(ProviderError):
    """Failed to create a host."""


class ImageNotFoundError(HostCreationError):
    """The specified image does not exist or is invalid."""

    def __init__(self, image: ImageReference) -> None:
        self.image = image
        super().__init__(f"Image not found: {image}")


class ResourceAllocationError(HostCreationError):
    """Failed to allocate resources for the host."""


class HostNameConflictError(ProviderError):
    """A host with this name already exists."""

    user_help_text = "Choose a different host name, or destroy the existing host first with 'mngr destroy'."

    def __init__(self, name: HostName) -> None:
        self.name = name
        super().__init__(f"Host name already exists: {name}")


class HostNotRunningError(ProviderError):
    """Host is not in RUNNING state."""

    user_help_text = "Start the host first with 'mngr start <host>'."

    def __init__(self, host_id: HostId, state: HostState) -> None:
        self.host_id = host_id
        self.state = state
        super().__init__(f"Host {host_id} is not running (state: {state})")


class HostNotStoppedError(ProviderError):
    """Host is not in STOPPED state."""

    user_help_text = "Stop the host first with 'mngr stop <host>'."

    def __init__(self, host_id: HostId, state: HostState) -> None:
        self.host_id = host_id
        self.state = state
        super().__init__(f"Host {host_id} is not stopped (state: {state})")


class SnapshotError(ProviderError):
    """Base class for snapshot-related errors."""


class SnapshotNotFoundError(SnapshotError):
    """No snapshot with this ID exists."""

    user_help_text = "Use 'mngr snapshot list <host>' to see available snapshots."

    def __init__(self, snapshot_id: SnapshotId) -> None:
        self.snapshot_id = snapshot_id
        super().__init__(f"Snapshot not found: {snapshot_id}")


class SnapshotsNotSupportedError(SnapshotError):
    """Provider does not support snapshots."""

    user_help_text = (
        "Snapshots are only available for cloud providers like Modal. The local provider does not support snapshots."
    )

    def __init__(self, provider_name: ProviderInstanceName) -> None:
        self.provider_name = provider_name
        super().__init__(f"Provider {provider_name} does not support snapshots")


class TagLimitExceededError(ProviderError):
    """Tags exceed provider's storage limit."""

    def __init__(self, limit: int, actual: int) -> None:
        self.limit = limit
        self.actual = actual
        super().__init__(f"Tag limit exceeded: {actual} tags (limit: {limit})")


class LocalHostNotStoppableError(ProviderError):
    """Raised when attempting to stop the local host."""

    def __init__(self) -> None:
        super().__init__("Cannot stop the local host - it is your local computer")


class LocalHostNotDestroyableError(ProviderError):
    """Raised when attempting to destroy the local host."""

    def __init__(self) -> None:
        super().__init__("Cannot destroy the local host - it is your local computer")


class PluginSpecifierError(BaseMngrError, ValueError):
    """Raised when a plugin specifier is invalid or cannot be resolved."""


class PluginMngrError(MngrError):
    """Raised when a plugin encounters an error during provisioning.

    Plugins should raise this error in the on_before_agent_provisioning hook
    when preconditions are not met (e.g., missing environment variables,
    missing required files).
    """


class ModalAuthError(PluginMngrError):
    """Modal authentication failed due to missing or invalid token."""

    def __init__(self) -> None:
        super().__init__(
            "Modal authentication failed. Token missing or invalid. "
            "You can disable the modal plugin by running "
            "'mngr config set --scope user plugins.modal.enabled false', "
            "or by passing --disable-plugin modal to individual commands. "
            "To configure modal credentials, see https://modal.com/docs/reference/modal.config"
        )


class ConfigError(MngrError):
    """Base class for config errors."""


class ConfigNotFoundError(ConfigError):
    """Config file not found."""


class ConfigParseError(ConfigError):
    """Failed to parse config file."""


class ConfigKeyNotFoundError(ConfigError, KeyError):
    """Configuration key not found."""

    def __init__(self, key: str) -> None:
        self.key = key
        super().__init__(f"Key not found: {key}")


class ConfigStructureError(ConfigError, TypeError):
    """Invalid configuration structure."""


class UnknownBackendError(ConfigError):
    """Unknown provider backend."""


class NestedTmuxError(MngrError):
    """Cannot attach to tmux session from inside another tmux session."""

    def __init__(self, session_name: str) -> None:
        self.session_name = session_name
        super().__init__(
            f"You're already in a tmux session. You can attach to the agent with:\n  tmux attach -t {session_name}"
        )
        self.user_help_text = (
            "To allow mngr to attach automatically inside tmux, run:\n"
            "  mngr config set --scope user is_nested_tmux_allowed true"
        )


class BinaryNotInstalledError(MngrError):
    """Raised when a required system binary is not installed."""

    def __init__(self, binary: str, purpose: str, install_hint: str) -> None:
        self.user_help_text = install_hint
        super().__init__(f"{binary} is required for {purpose} but was not found on PATH")



from imbue.mngr.config.data_types import PluginConfig
from imbue.mngr.primitives import PluginName

# =============================================================================
# Plugin Config Registry
# =============================================================================

_plugin_config_registry: dict[PluginName, type[PluginConfig]] = {}


def register_plugin_config(
    plugin_name: str,
    config_class: type[PluginConfig],
) -> None:
    """Register a plugin config class for a plugin."""
    _plugin_config_registry[PluginName(plugin_name)] = config_class


def get_plugin_config_class(plugin_name: str) -> type[PluginConfig]:
    """Get the config class for a plugin.

    Returns the base PluginConfig if no specific type is registered.
    """
    key = PluginName(plugin_name)
    if key not in _plugin_config_registry:
        return PluginConfig
    return _plugin_config_registry[key]


def list_registered_plugins() -> list[str]:
    """List all registered plugin names."""
    return sorted(str(k) for k in _plugin_config_registry.keys())



from collections.abc import Sequence
from typing import Any
from typing import assert_never

import click
from click_option_group import optgroup
from loguru import logger

from imbue.imbue_common.pure import pure
from imbue.mngr.api.find import AgentMatch
from imbue.mngr.api.find import find_agents_by_identifiers_or_state
from imbue.mngr.api.find import group_agents_by_host
from imbue.mngr.api.find import resolve_host_reference
from imbue.mngr.api.list import load_all_agents_grouped_by_host
from imbue.mngr.api.providers import get_provider_instance
from imbue.mngr.cli.common_opts import CommonCliOptions
from imbue.mngr.cli.common_opts import add_common_options
from imbue.mngr.cli.common_opts import setup_command_context
from imbue.mngr.cli.help_formatter import CommandHelpMetadata
from imbue.mngr.cli.help_formatter import add_pager_help_option
from imbue.mngr.cli.help_formatter import register_help_metadata
from imbue.mngr.cli.output_helpers import emit_event
from imbue.mngr.cli.output_helpers import emit_final_json
from imbue.mngr.cli.output_helpers import write_human_line
from imbue.mngr.config.data_types import MngrContext
from imbue.mngr.config.data_types import OutputOptions
from imbue.mngr.errors import AgentNotFoundOnHostError
from imbue.mngr.errors import HostOfflineError
from imbue.mngr.hosts.common import get_activity_sources_for_idle_mode
from imbue.mngr.interfaces.data_types import ActivityConfig
from imbue.mngr.interfaces.host import HostInterface
from imbue.mngr.interfaces.host import OnlineHostInterface
from imbue.mngr.primitives import ActivitySource
from imbue.mngr.primitives import HostId
from imbue.mngr.primitives import HostReference
from imbue.mngr.primitives import IdleMode
from imbue.mngr.primitives import OutputFormat
from imbue.mngr.primitives import Permission
from imbue.mngr.utils.duration import parse_duration_to_seconds


class LimitCliOptions(CommonCliOptions):
    """Options passed from the CLI to the limit command."""

    agents: tuple[str, ...]
    agent_list: tuple[str, ...]
    hosts: tuple[str, ...]
    limit_all: bool
    dry_run: bool
    # Planned features (not yet implemented)
    include: tuple[str, ...]
    exclude: tuple[str, ...]
    stdin: bool
    # Lifecycle
    start_on_boot: bool | None
    idle_timeout: str | None
    idle_mode: str | None
    activity_sources: str | None
    add_activity_source: tuple[str, ...]
    remove_activity_source: tuple[str, ...]
    # Permissions
    grant: tuple[str, ...]
    revoke: tuple[str, ...]
    # SSH Keys (not yet implemented)
    refresh_ssh_keys: bool
    add_ssh_key: tuple[str, ...]
    remove_ssh_key: tuple[str, ...]


def _make_idle_mode_choices() -> list[str]:
    """Get lowercase idle mode choices (excluding CUSTOM, which is derived, not user-settable)."""
    return [m.value.lower() for m in IdleMode if m != IdleMode.CUSTOM]


def _make_activity_source_choices() -> list[str]:
    """Get lowercase activity source choices."""
    return [s.value.lower() for s in ActivitySource]


def _output(message: str, output_opts: OutputOptions) -> None:
    """Output a message according to the format."""
    if output_opts.output_format == OutputFormat.HUMAN:
        write_human_line(message)


def _output_result(
    changes: list[dict[str, Any]],
    output_opts: OutputOptions,
) -> None:
    """Output the final result."""
    result_data = {"changes": changes, "count": len(changes)}
    match output_opts.output_format:
        case OutputFormat.JSON:
            emit_final_json(result_data)
        case OutputFormat.JSONL:
            emit_event("limit_result", result_data, OutputFormat.JSONL)
        case OutputFormat.HUMAN:
            if changes:
                write_human_line("Applied {} change(s)", len(changes))
        case _ as unreachable:
            assert_never(unreachable)


@pure
def _build_updated_activity_config(
    current: ActivityConfig,
    idle_timeout_str: str | None,
    idle_mode_str: str | None,
    activity_sources_str: str | None,
    add_activity_source: tuple[str, ...],
    remove_activity_source: tuple[str, ...],
) -> ActivityConfig:
    """Build an updated ActivityConfig by merging current config with requested changes.

    idle_mode is a computed property on ActivityConfig (derived from activity_sources),
    so when --idle-mode is specified we convert it to the corresponding activity sources
    via get_activity_sources_for_idle_mode.
    """
    new_idle_timeout = (
        int(parse_duration_to_seconds(idle_timeout_str))
        if idle_timeout_str is not None
        else current.idle_timeout_seconds
    )

    if activity_sources_str is not None:
        # Explicit --activity-sources replaces everything
        new_activity_sources = tuple(ActivitySource(s.strip().upper()) for s in activity_sources_str.split(","))
    elif idle_mode_str is not None:
        # --idle-mode sets the canonical activity sources for that mode
        new_activity_sources = get_activity_sources_for_idle_mode(IdleMode(idle_mode_str.upper()))
    else:
        # Incremental changes via --add/--remove-activity-source
        current_sources = set(current.activity_sources)
        for source_str in add_activity_source:
            current_sources.add(ActivitySource(source_str.upper()))
        for source_str in remove_activity_source:
            current_sources.discard(ActivitySource(source_str.upper()))
        new_activity_sources = tuple(current_sources)

    return ActivityConfig(
        idle_timeout_seconds=new_idle_timeout,
        activity_sources=new_activity_sources,
    )


@pure
def _build_updated_permissions(
    current: Sequence[Permission],
    grant: tuple[str, ...],
    revoke: tuple[str, ...],
) -> list[Permission]:
    """Build an updated permissions list by applying grants and revokes."""
    result = set(current)
    for perm_str in grant:
        result.add(Permission(perm_str))
    for perm_str in revoke:
        result.discard(Permission(perm_str))
    return sorted(result)


def _has_host_level_settings(opts: LimitCliOptions) -> bool:
    """Return True if any host-level settings are being changed."""
    return (
        opts.idle_timeout is not None
        or opts.idle_mode is not None
        or opts.activity_sources is not None
        or len(opts.add_activity_source) > 0
        or len(opts.remove_activity_source) > 0
    )


def _has_agent_level_settings(opts: LimitCliOptions) -> bool:
    """Return True if any agent-level settings are being changed."""
    return opts.start_on_boot is not None or len(opts.grant) > 0 or len(opts.revoke) > 0


def _has_any_setting(opts: LimitCliOptions) -> bool:
    """Return True if any setting is being changed."""
    return _has_host_level_settings(opts) or _has_agent_level_settings(opts)


def _apply_activity_config_to_host(
    online_host: OnlineHostInterface,
    host_id_str: str,
    opts: LimitCliOptions,
    output_opts: OutputOptions,
    changes: list[dict[str, Any]],
) -> None:
    """Apply activity config changes to a single online host."""
    current_config = online_host.get_activity_config()
    new_config = _build_updated_activity_config(
        current=current_config,
        idle_timeout_str=opts.idle_timeout,
        idle_mode_str=opts.idle_mode,
        activity_sources_str=opts.activity_sources,
        add_activity_source=opts.add_activity_source,
        remove_activity_source=opts.remove_activity_source,
    )
    online_host.set_activity_config(new_config)
    _output(f"Updated activity config for host {host_id_str}", output_opts)
    changes.append(
        {
            "type": "host_activity_config",
            "host_id": host_id_str,
        }
    )


def _build_host_references(mngr_ctx: MngrContext) -> list[HostReference]:
    """Build a deduplicated list of HostReferences from all known agents."""
    agents_by_host, _ = load_all_agents_grouped_by_host(mngr_ctx, include_destroyed=False)
    return list(agents_by_host.keys())


def _resolve_host_identifiers(
    host_identifiers: tuple[str, ...],
    mngr_ctx: MngrContext,
) -> set[HostId]:
    """Resolve host identifiers (names or IDs) to a set of HostIds.

    Raises UserInputError if any host identifier cannot be resolved.
    """
    all_hosts = _build_host_references(mngr_ctx)
    resolved_ids: set[HostId] = set()
    for host_identifier in host_identifiers:
        # resolve_host_reference raises UserInputError for unresolvable identifiers;
        # it only returns None when host_identifier is None, which cannot happen here
        resolved_host = resolve_host_reference(host_identifier, all_hosts)
        assert resolved_host is not None
        resolved_ids.add(resolved_host.host_id)
    return resolved_ids


@click.command(name="limit")
@click.argument("agents", nargs=-1, required=False)
@optgroup.group("Target Selection")
@optgroup.option(
    "--agent",
    "agent_list",
    multiple=True,
    help="Agent name or ID to configure (can be specified multiple times)",
)
@optgroup.option(
    "--host",
    "hosts",
    multiple=True,
    help="Host name or ID to configure (can be specified multiple times)",
)
@optgroup.option(
    "-a",
    "--all",
    "--all-agents",
    "limit_all",
    is_flag=True,
    help="Apply limits to all agents",
)
@optgroup.option(
    "--include",
    multiple=True,
    help="Filter agents to configure by CEL expression (repeatable) [future]",
)
@optgroup.option(
    "--exclude",
    multiple=True,
    help="Exclude agents matching CEL expression (repeatable) [future]",
)
@optgroup.option(
    "--stdin",
    is_flag=True,
    help="Read agent and host names/IDs from stdin, one per line [future]",
)
@optgroup.group("Behavior")
@optgroup.option(
    "--dry-run",
    is_flag=True,
    help="Show what limits would be changed without actually changing them",
)
@optgroup.group("Lifecycle")
@optgroup.option(
    "--start-on-boot/--no-start-on-boot",
    default=None,
    help="Automatically restart agent when host restarts",
)
@optgroup.option(
    "--idle-timeout",
    type=str,
    default=None,
    help="Shutdown after idle for specified duration (e.g., 30s, 5m, 1h, or plain seconds)",
)
@optgroup.option(
    "--idle-mode",
    type=click.Choice(_make_idle_mode_choices(), case_sensitive=False),
    default=None,
    help="When to consider host idle",
)
@optgroup.option(
    "--activity-sources",
    type=str,
    default=None,
    help="Set activity sources for idle detection (comma-separated)",
)
@optgroup.option(
    "--add-activity-source",
    type=click.Choice(_make_activity_source_choices(), case_sensitive=False),
    multiple=True,
    help="Add an activity source for idle detection (repeatable)",
)
@optgroup.option(
    "--remove-activity-source",
    type=click.Choice(_make_activity_source_choices(), case_sensitive=False),
    multiple=True,
    help="Remove an activity source from idle detection (repeatable)",
)
@optgroup.group("Permissions")
@optgroup.option(
    "--grant",
    multiple=True,
    help="Grant a permission to the agent (repeatable)",
)
@optgroup.option(
    "--revoke",
    multiple=True,
    help="Revoke a permission from the agent (repeatable)",
)
@optgroup.group("SSH Keys")
@optgroup.option(
    "--refresh-ssh-keys",
    is_flag=True,
    help="Refresh the SSH keys for the host [future]",
)
@optgroup.option(
    "--add-ssh-key",
    multiple=True,
    help="Add an SSH public key to the host for access (repeatable) [future]",
)
@optgroup.option(
    "--remove-ssh-key",
    multiple=True,
    help="Remove an SSH public key from the host (repeatable) [future]",
)
@add_common_options
@click.pass_context
def limit(ctx: click.Context, **kwargs: Any) -> None:
    """Configure limits for agents and hosts. [experimental]

    Configures settings on existing agents and hosts: idle timeout, idle mode,
    activity sources, permissions, and start-on-boot.

    \b
    Alias: lim

    \b
    Examples:

      mngr limit my-agent --idle-timeout 5m

      mngr limit my-agent --grant network --grant internet

      mngr limit --all --idle-mode disabled

      mngr limit --host my-host --idle-timeout 1h
    """
    mngr_ctx, output_opts, opts = setup_command_context(
        ctx=ctx,
        command_name="limit",
        command_class=LimitCliOptions,
    )
    logger.debug("Started limit command")

    # Check for unsupported [future] options
    if opts.include:
        raise NotImplementedError("--include is not implemented yet")
    if opts.exclude:
        raise NotImplementedError("--exclude is not implemented yet")
    if opts.stdin:
        raise NotImplementedError("--stdin is not implemented yet")
    if opts.refresh_ssh_keys:
        raise NotImplementedError("--refresh-ssh-keys is not implemented yet")
    if opts.add_ssh_key:
        raise NotImplementedError("--add-ssh-key is not implemented yet")
    if opts.remove_ssh_key:
        raise NotImplementedError("--remove-ssh-key is not implemented yet")

    # Validate at least one setting is being changed
    if not _has_any_setting(opts):
        raise click.UsageError(
            "Must specify at least one setting to change (e.g., --idle-timeout, --idle-mode, "
            "--activity-sources, --start-on-boot, --grant, --revoke)"
        )

    # Validate --activity-sources is not combined with --add/--remove-activity-source
    if opts.activity_sources is not None and (opts.add_activity_source or opts.remove_activity_source):
        raise click.UsageError(
            "Cannot combine --activity-sources with --add-activity-source or --remove-activity-source"
        )

    # Validate targets: must specify agents, --host, or --all
    agent_identifiers = list(opts.agents) + list(opts.agent_list)
    has_agents = bool(agent_identifiers)
    has_hosts = bool(opts.hosts)

    if not has_agents and not has_hosts and not opts.limit_all:
        raise click.UsageError("Must specify at least one agent, --host, or --all")

    if has_agents and opts.limit_all:
        raise click.UsageError("Cannot specify both agent names and --all")

    # If only --host is specified (no agents), agent-level settings are not allowed
    if has_hosts and not has_agents and not opts.limit_all and _has_agent_level_settings(opts):
        raise click.UsageError(
            "Agent-level settings (--start-on-boot, --grant, --revoke) require agent targeting. "
            "Use --agent or --all with --host to target agents on specific hosts."
        )

    # If --host only (no agents, no --all), apply host-level changes directly
    if has_hosts and not has_agents and not opts.limit_all:
        changes: list[dict[str, Any]] = []
        all_hosts = _build_host_references(mngr_ctx)
        for host_identifier in opts.hosts:
            _apply_host_only_changes(
                host_identifier=host_identifier,
                all_hosts=all_hosts,
                opts=opts,
                output_opts=output_opts,
                mngr_ctx=mngr_ctx,
                dry_run=opts.dry_run,
                changes=changes,
            )
        _output_result(changes, output_opts)
        return

    # Find agents (match all states for limit command)
    agents = find_agents_by_identifiers_or_state(
        agent_identifiers=agent_identifiers,
        filter_all=opts.limit_all,
        target_state=None,
        mngr_ctx=mngr_ctx,
    )

    if not agents:
        _output("No agents found to configure", output_opts)
        return

    # If --host is also specified, filter agents to those on the specified hosts
    if has_hosts:
        resolved_host_ids = _resolve_host_identifiers(opts.hosts, mngr_ctx)
        target_agents = [a for a in agents if a.host_id in resolved_host_ids]
        if not target_agents:
            _output("No agents found on the specified host(s)", output_opts)
            return
    else:
        target_agents = agents

    # Handle dry-run mode
    if opts.dry_run:
        _output("Would configure:", output_opts)
        for match in target_agents:
            _output(f"  - {match.agent_name} (on host {match.host_id})", output_opts)
        if _has_host_level_settings(opts):
            unique_hosts = {str(m.host_id) for m in target_agents}
            _output(f"Host-level changes would apply to {len(unique_hosts)} host(s)", output_opts)
        return

    # Apply changes
    changes = []
    agents_by_host = group_agents_by_host(target_agents)
    updated_host_ids: set[str] = set()
    has_permission_changes = bool(opts.grant or opts.revoke)

    for host_key, agent_list in agents_by_host.items():
        host_id_str, _ = host_key.split(":", 1)
        provider_name = agent_list[0].provider_name

        provider = get_provider_instance(provider_name, mngr_ctx)
        host = provider.get_host(HostId(host_id_str))

        match host:
            case OnlineHostInterface() as online_host:
                # Apply host-level changes once per host
                if _has_host_level_settings(opts) and host_id_str not in updated_host_ids:
                    _apply_activity_config_to_host(
                        online_host=online_host,
                        host_id_str=host_id_str,
                        opts=opts,
                        output_opts=output_opts,
                        changes=changes,
                    )
                    updated_host_ids.add(host_id_str)

                # Apply agent-level changes per agent
                if _has_agent_level_settings(opts):
                    for agent_match in agent_list:
                        _apply_agent_changes(
                            agent_match=agent_match,
                            online_host=online_host,
                            opts=opts,
                            output_opts=output_opts,
                            changes=changes,
                        )

            case HostInterface():
                raise HostOfflineError(f"Host '{host_id_str}' is offline. Cannot configure agents on offline hosts.")
            case _ as unreachable:
                assert_never(unreachable)

    if has_permission_changes:
        _output("Restart required for permission changes to take effect.", output_opts)

    _output_result(changes, output_opts)


def _apply_host_only_changes(
    host_identifier: str,
    all_hosts: list[HostReference],
    opts: LimitCliOptions,
    output_opts: OutputOptions,
    dry_run: bool,
    changes: list[dict[str, Any]],
    mngr_ctx: MngrContext,
) -> None:
    """Apply host-level changes when targeting hosts directly (no agents).

    Raises UserInputError if the host identifier cannot be resolved.
    """
    # resolve_host_reference raises UserInputError for unresolvable identifiers;
    # it only returns None when host_identifier is None, which cannot happen here
    resolved_host = resolve_host_reference(host_identifier, all_hosts)
    assert resolved_host is not None

    if dry_run:
        _output(f"Would update activity config for host {resolved_host.host_id}", output_opts)
        return

    provider = get_provider_instance(resolved_host.provider_name, mngr_ctx)
    host = provider.get_host(resolved_host.host_id)

    match host:
        case OnlineHostInterface() as online_host:
            _apply_activity_config_to_host(
                online_host=online_host,
                host_id_str=str(resolved_host.host_id),
                opts=opts,
                output_opts=output_opts,
                changes=changes,
            )
        case HostInterface():
            raise HostOfflineError(f"Host '{resolved_host.host_id}' is offline. Cannot configure offline hosts.")
        case _ as unreachable:
            assert_never(unreachable)


def _apply_agent_changes(
    agent_match: AgentMatch,
    online_host: OnlineHostInterface,
    opts: LimitCliOptions,
    output_opts: OutputOptions,
    changes: list[dict[str, Any]],
) -> None:
    """Apply agent-level changes to a single agent."""
    for agent in online_host.get_agents():
        if agent.id == agent_match.agent_id:
            if opts.start_on_boot is not None:
                agent.set_is_start_on_boot(opts.start_on_boot)
                _output(
                    f"Set start-on-boot={opts.start_on_boot} for agent {agent_match.agent_name}",
                    output_opts,
                )
                changes.append(
                    {
                        "type": "agent_start_on_boot",
                        "agent_id": str(agent_match.agent_id),
                        "agent_name": str(agent_match.agent_name),
                        "start_on_boot": opts.start_on_boot,
                    }
                )

            if opts.grant or opts.revoke:
                current_permissions = agent.get_permissions()
                new_permissions = _build_updated_permissions(
                    current=current_permissions,
                    grant=opts.grant,
                    revoke=opts.revoke,
                )
                agent.set_permissions(new_permissions)
                _output(
                    f"Updated permissions for agent {agent_match.agent_name}",
                    output_opts,
                )
                changes.append(
                    {
                        "type": "agent_permissions",
                        "agent_id": str(agent_match.agent_id),
                        "agent_name": str(agent_match.agent_name),
                        "permissions": [str(p) for p in new_permissions],
                    }
                )
            break
    else:
        raise AgentNotFoundOnHostError(agent_match.agent_id, agent_match.host_id)


# Register help metadata for git-style help formatting
_LIMIT_HELP_METADATA = CommandHelpMetadata(
    name="mngr-limit",
    one_line_description="Configure limits for agents and hosts [experimental]",
    synopsis="mngr [limit|lim] [AGENTS...] [--agent <AGENT>] [--host <HOST>] [--all] [--idle-timeout <DURATION>] [--idle-mode <MODE>] [--grant <PERM>] [--revoke <PERM>]",
    arguments_description="- `AGENTS`: Agent name(s) or ID(s) to configure (can also be specified via `--agent`)",
    description="""Configure settings on existing agents and hosts: idle timeout,
idle mode, activity sources, permissions, and start-on-boot.

Agents effectively have permissions that are equivalent to the *union* of all
permissions on the same host. Changing permissions for agents requires them
to be restarted.

Changes to some limits for hosts (e.g. CPU, RAM, disk space, network) are
handled by the provider.

When targeting agents, host-level settings (idle-timeout, idle-mode,
activity-sources) are applied to each agent's underlying host.

Agent-level settings (start-on-boot, grant, revoke) require agent targeting
and cannot be used with --host alone.""",
    aliases=("lim",),
    examples=(
        ("Set idle timeout for an agent's host", "mngr limit my-agent --idle-timeout 5m"),
        ("Grant permissions to an agent", "mngr limit my-agent --grant network --grant internet"),
        ("Disable idle detection for all agents", "mngr limit --all --idle-mode disabled"),
        ("Update host idle settings directly", "mngr limit --host my-host --idle-timeout 1h"),
        ("Preview changes without applying", "mngr limit --all --idle-timeout 5m --dry-run"),
    ),
    see_also=(
        ("create", "Create a new agent"),
        ("list", "List existing agents"),
        ("stop", "Stop running agents"),
    ),
    additional_sections=(
        (
            "Idle Modes",
            "See [Idle Detection](../../concepts/idle_detection.md) for details on idle modes and activity sources.",
        ),
    ),
)

register_help_metadata("limit", _LIMIT_HELP_METADATA)
for alias in _LIMIT_HELP_METADATA.aliases:
    register_help_metadata(alias, _LIMIT_HELP_METADATA)

add_pager_help_option(limit)



from collections.abc import Sequence
from typing import Any
from typing import assert_never

import click
from click_option_group import optgroup
from loguru import logger

from imbue.imbue_common.logging import log_span
from imbue.mngr.api.connect import connect_to_agent
from imbue.mngr.api.connect import resolve_connect_command
from imbue.mngr.api.connect import run_connect_command
from imbue.mngr.api.data_types import ConnectionOptions
from imbue.mngr.api.find import ensure_host_started
from imbue.mngr.api.find import find_agents_by_identifiers_or_state
from imbue.mngr.api.find import group_agents_by_host
from imbue.mngr.api.providers import get_provider_instance
from imbue.mngr.cli.common_opts import CommonCliOptions
from imbue.mngr.cli.common_opts import add_common_options
from imbue.mngr.cli.common_opts import setup_command_context
from imbue.mngr.cli.completion import complete_agent_name
from imbue.mngr.cli.help_formatter import CommandHelpMetadata
from imbue.mngr.cli.help_formatter import add_pager_help_option
from imbue.mngr.cli.help_formatter import register_help_metadata
from imbue.mngr.cli.output_helpers import emit_event
from imbue.mngr.cli.output_helpers import emit_final_json
from imbue.mngr.cli.output_helpers import emit_format_template_lines
from imbue.mngr.cli.output_helpers import write_human_line
from imbue.mngr.config.data_types import OutputOptions
from imbue.mngr.interfaces.agent import AgentInterface
from imbue.mngr.primitives import AgentLifecycleState
from imbue.mngr.primitives import HostId
from imbue.mngr.primitives import OutputFormat
from imbue.mngr.utils.polling import poll_until


class StartCliOptions(CommonCliOptions):
    """Options passed from the CLI to the start command."""

    agents: tuple[str, ...]
    agent_list: tuple[str, ...]
    start_all: bool
    dry_run: bool
    connect: bool
    connect_command: str | None
    # Planned features (not yet implemented)
    host: tuple[str, ...]
    include: tuple[str, ...]
    exclude: tuple[str, ...]
    stdin: bool
    snapshot: str | None
    latest: bool


def _output(message: str, output_opts: OutputOptions) -> None:
    """Output a message according to the format."""
    if output_opts.output_format == OutputFormat.HUMAN:
        write_human_line(message)


def _output_result(started_agents: Sequence[str], output_opts: OutputOptions) -> None:
    """Output the final result."""
    if output_opts.format_template is not None:
        items = [{"name": name} for name in started_agents]
        emit_format_template_lines(output_opts.format_template, items)
        return
    result_data = {"started_agents": started_agents, "count": len(started_agents)}
    match output_opts.output_format:
        case OutputFormat.JSON:
            emit_final_json(result_data)
        case OutputFormat.JSONL:
            emit_event("start_result", result_data, OutputFormat.JSONL)
        case OutputFormat.HUMAN:
            if started_agents:
                write_human_line("Successfully started {} agent(s)", len(started_agents))
        case _ as unreachable:
            assert_never(unreachable)


def _send_resume_message_if_configured(agent: AgentInterface, output_opts: OutputOptions) -> None:
    """Send the resume message to an agent if one is configured."""
    resume_message = agent.get_resume_message()
    if resume_message is None:
        return

    _output(f"Sending resume message to {agent.name}...", output_opts)
    # Wait for the agent to signal readiness via the WAITING lifecycle state.
    # Agents like Claude configure hooks that remove the 'active' file when idle.
    # If the timeout expires (agent doesn't support hooks or is slow), proceed anyway.
    timeout = agent.get_ready_timeout_seconds()
    with log_span("Waiting for agent to become ready before sending resume message"):
        is_ready = poll_until(
            lambda: agent.get_lifecycle_state() == AgentLifecycleState.WAITING,
            timeout=timeout,
            poll_interval=0.2,
        )
    if is_ready:
        logger.debug("Signaled agent readiness via WAITING state")
    else:
        logger.debug(
            "Failed to reach WAITING state within {}s, proceeding anyway",
            timeout,
        )
    agent.send_message(resume_message)
    logger.debug("Sent resume message to agent {}", agent.name)


@click.command(name="start")
@click.argument("agents", nargs=-1, required=False, shell_complete=complete_agent_name)
@optgroup.group("Target Selection")
@optgroup.option(
    "--agent",
    "agent_list",
    multiple=True,
    help="Agent name or ID to start (can be specified multiple times)",
)
@optgroup.option(
    "-a",
    "--all",
    "--all-agents",
    "start_all",
    is_flag=True,
    help="Start all stopped agents",
)
@optgroup.option(
    "--host",
    multiple=True,
    help="Host(s) to start all stopped agents on [repeatable] [future]",
)
@optgroup.option(
    "--include",
    multiple=True,
    help="Filter agents and hosts to start by CEL expression (repeatable) [future]",
)
@optgroup.option(
    "--exclude",
    multiple=True,
    help="Exclude agents and hosts matching CEL expression (repeatable) [future]",
)
@optgroup.option(
    "--stdin",
    is_flag=True,
    help="Read agent and host names/IDs from stdin, one per line [future]",
)
@optgroup.group("Behavior")
@optgroup.option(
    "--dry-run",
    is_flag=True,
    help="Show what would be started without actually starting",
)
@optgroup.option(
    "--connect/--no-connect",
    default=False,
    help="Connect to the agent after starting (only valid for single agent)",
)
@optgroup.option(
    "--connect-command",
    help="Command to run instead of the builtin connect. MNGR_AGENT_NAME and MNGR_SESSION_NAME env vars are set.",
)
@optgroup.group("Snapshot")
@optgroup.option(
    "--snapshot",
    type=str,
    default=None,
    help="Start from a specific snapshot instead of the most recent [future]",
)
@optgroup.option(
    "--latest/--no-latest",
    default=True,
    help="Start from the most recent snapshot or state [default] [future]",
)
@add_common_options
@click.pass_context
def start(ctx: click.Context, **kwargs: Any) -> None:
    """Start stopped agent(s).

    For remote hosts, this restores from the most recent snapshot and starts
    the container/instance. For local agents, this starts the agent's tmux
    session.

    Supports custom format templates via --format. Available fields: name.

    \b
    Examples:

      mngr start my-agent

      mngr start agent1 agent2

      mngr start --agent my-agent --connect

      mngr start --all

      mngr start --all --format '{name}'
    """
    mngr_ctx, output_opts, opts = setup_command_context(
        ctx=ctx,
        command_name="start",
        command_class=StartCliOptions,
        is_format_template_supported=True,
    )

    # Check for unsupported [future] options
    if opts.host:
        raise NotImplementedError("--host is not implemented yet")
    if opts.include:
        raise NotImplementedError("--include is not implemented yet")
    if opts.exclude:
        raise NotImplementedError("--exclude is not implemented yet")
    if opts.stdin:
        raise NotImplementedError("--stdin is not implemented yet")
    if opts.snapshot is not None:
        raise NotImplementedError("--snapshot is not implemented yet")
    if not opts.latest:
        raise NotImplementedError("--no-latest is not implemented yet")

    # Validate input
    agent_identifiers = list(opts.agents) + list(opts.agent_list)

    if not agent_identifiers and not opts.start_all:
        raise click.UsageError("Must specify at least one agent or use --all")

    if agent_identifiers and opts.start_all:
        raise click.UsageError("Cannot specify both agent names and --all")

    if opts.connect and (opts.start_all or len(agent_identifiers) > 1):
        raise click.UsageError("--connect can only be used with a single agent")

    # Find agents to start (STOPPED agents when using --all)
    agents_to_start = find_agents_by_identifiers_or_state(
        agent_identifiers=agent_identifiers,
        filter_all=opts.start_all,
        target_state=AgentLifecycleState.STOPPED,
        mngr_ctx=mngr_ctx,
    )

    if not agents_to_start:
        _output("No stopped agents found to start", output_opts)
        return

    # Handle dry-run mode
    if opts.dry_run:
        _output("Would start:", output_opts)
        for match in agents_to_start:
            _output(f"  - {match.agent_name} (on host {match.host_id})", output_opts)
        return

    # Start each agent
    started_agents: list[str] = []
    last_started_agent = None
    last_started_host = None

    # Group agents by host to avoid starting the same host multiple times
    agents_by_host = group_agents_by_host(agents_to_start)

    for host_key, agent_list in agents_by_host.items():
        host_id_str, _ = host_key.split(":", 1)
        # Get provider from first agent (all agents in list have same provider)
        provider_name = agent_list[0].provider_name

        provider = get_provider_instance(provider_name, mngr_ctx)
        host = provider.get_host(HostId(host_id_str))

        # Ensure host is started (always start since this is the start command)
        online_host, _ = ensure_host_started(host, is_start_desired=True, provider=provider)

        # Start each agent on this host
        agent_ids_to_start = [match.agent_id for match in agent_list]
        online_host.start_agents(agent_ids_to_start)

        for match in agent_list:
            started_agents.append(str(match.agent_name))
            _output(f"Started agent: {match.agent_name}", output_opts)

            # Get the agent object for potential connect and resume message
            for agent in online_host.get_agents():
                if agent.id == match.agent_id:
                    # Send resume message if configured
                    _send_resume_message_if_configured(agent, output_opts)

                    # Track for potential connect
                    if opts.connect:
                        last_started_agent = agent
                        last_started_host = online_host
                    break

    # Output final result
    _output_result(started_agents, output_opts)

    # Connect if requested and we started exactly one agent
    if opts.connect and last_started_agent is not None and last_started_host is not None:
        resolved_command = resolve_connect_command(opts.connect_command, mngr_ctx)
        if resolved_command is not None:
            session_name = f"{mngr_ctx.config.prefix}{last_started_agent.name}"
            run_connect_command(
                resolved_command,
                str(last_started_agent.name),
                session_name,
                is_local=last_started_host.is_local,
            )
        else:
            connection_opts = ConnectionOptions(
                is_reconnect=True,
                retry_count=3,
                retry_delay="5s",
                attach_command=None,
                is_unknown_host_allowed=False,
            )
            logger.info("Connecting to agent: {}", last_started_agent.name)
            connect_to_agent(last_started_agent, last_started_host, mngr_ctx, connection_opts)


# Register help metadata for git-style help formatting
_START_HELP_METADATA = CommandHelpMetadata(
    name="mngr-start",
    one_line_description="Start stopped agent(s)",
    synopsis="mngr start [AGENTS...] [--agent <AGENT>] [--all] [--host <HOST>] [--connect] [--dry-run] [--snapshot <SNAPSHOT>]",
    description="""Start one or more stopped agents.

For remote hosts, this restores from the most recent snapshot and starts
the container/instance. For local agents, this starts the agent's tmux
session.

If multiple agents share a host, they will all be started together when
the host starts.""",
    aliases=(),
    examples=(
        ("Start an agent by name", "mngr start my-agent"),
        ("Start multiple agents", "mngr start agent1 agent2"),
        ("Start and connect", "mngr start my-agent --connect"),
        ("Start all stopped agents", "mngr start --all"),
        ("Preview what would be started", "mngr start --all --dry-run"),
    ),
    see_also=(
        ("stop", "Stop running agents"),
        ("connect", "Connect to an agent"),
        ("list", "List existing agents"),
    ),
)

register_help_metadata("start", _START_HELP_METADATA)

# Add pager-enabled help option to the start command
add_pager_help_option(start)
